/****************************************************
;;; MODULE NAME : pars.y
;;;
;;; PARENT      : main.c
;;;
;;; PROJECT     : MPLOP
;;;
;;; SCCS        : %W%  %G%
;;;
;;; AUTHOR      : Janos Breiteneicher
;;;
;;; DATE        : 15.5.91
;;;
;;; DESCRIPTION : YACC grammar for mplop
;;;
;;; REMARKS     :
;;;		3.3.93	jsc	parsed_rule_type added
;;;		2.7.93	jsc	extern.h
;;;		14.12.93	jsc		fix of
;;;						bug ''ishorn'':
;;;						Queries contain, by default
;;;						one positive literal
;;;		3012.93		hamdi		ddepth,dinf
;;;		24.3.94		jsc		pred-names also on 	
;;;						term position
;;;		6.4.94		jsc		yyerror into pars.y
;;;						from yaccsed.script
;;;		21.6.95		jsc,klaus	fcb
;;;
;;;
****************************************************/

%token LARR COLMIN QUERY		/* <- :- ?- 		*/
%token ASSIGN EVASSIGN IS RELOP		/* := :is is < >	*/
%token NAME GLOBVAR VAR STRING NUMBER
%token BEMOD BECOM BECON BEIMP
%token ENMOD ENCOM ENCON ENIMP
%token EXPO IMPO GLOB 			/* interface  	*/
%token COPYS 				/* $Copies 	*/
%token CONSTRSEP CONSTRAIN STRUCVAR 	/* constraints	*/
%token '.' ';' ',' '|' '~'		/* single chars	*/
%token '[' ']' '(' ')'  

%left '+' '-'  				/* priority levels */
%left '*' '/'

%{
#include <stdio.h>
#include <ctype.h>
#include "types.h"
#include "extern.h"

#define NULLTERM (termtype *)NULL
#define NULLLIST (symbollist *)NULL

/******************************************************************************
/* type of parser stack 
/*****************************************************************************/
typedef union {
		struct mocl {
		 modtype * modu;
		 claustype * claus;
		}		MC;
		claustype *	C;
		predtype *	P;
		termtype *	T;
		symbollist * 	S;
		intertype 	I;
		int		N;
	} YYSTYPE;

/******************************************************************************
/* statics
/*****************************************************************************/
static int	numval;		/* value delivered by scanner */
static int	modcnt	 = 1;	/* module counter */
static int	modunr   = 0;	/* current number of module */
static int      clausenr = 1;   /* current number of clause */
static int	litnr	 = 1;	/* current number of literal in clause */
static int	builtnr	 = 1;	/* current number of built_in in clause */
static int	poslitnr = 0;	/* number of positive literals in clause */
static int	aktsign = '+';	/* sign '-' only in tails */
static int	termsize;
static int	termdepth;

static modtype    *currmodule = NULL;	/* pointer to static predecessor */
static claustype  *currclause = NULL;	/* pointer to current clause	*/
static predtype   *pp;
static termtype   *tp;
static symbollist *slp;

#define MAXDIFFVARS	100

static int	varlist[MAXDIFFVARS];
static int	*varlist_ptr;

#define VARUSAGE(T)	\
	if (!(symbtab[T->symb].index)){ \
		*(varlist_ptr++) = T->symb; \
		symbtab[T->symb].index=varlist_ptr-varlist; \
		} 

%}

%type <MC> body modules module
%type <C>  clause strategy dynstrat stastrat 
%type <P>  binexpr literal predicate head tail synclause controling
%type <P>  opt_constraints constraints constraint
%type <T>  numexpr term terms glvar list conslist
%type <I>  interface 
%type <S>  exports imports globals
%type <N>  GLOBVAR RELOP NAME 
%type <N>  '+' '-' '*' '/'

%start mploprogram

%%

	mploprogram :
		 { 
		 init_symb(); /* initialize symboltable, enter built-ins */
		 matrix = init_mod(enter_token("main_"));
		 }
		strategy modules
		 {
		 if (errors) {
		   fprintf(stderr,"parsing: %d errors occurred\n",errors);
		   return(1);
		   }
		 matrix->strategy= $<C>2;
		 matrix->m_list	 = $<MC>3.modu;
		 matrix->c_list  = $<MC>3.claus;
		    /* default query is standard */
		 prefix_query(matrix);
		 enter_interface(matrix);
		 nr_clauses = clausenr-1;	/* save in global variable */
		 return(0);
		 }
		;

	modules :
		module 
		| module modules
		 {
		 if ($<MC>1.modu) {
		    $<MC>1.modu->next = $<MC>2.modu;
		    $<MC>1.claus = $<MC>2.claus;
		    }
		 else {
		    $<MC>1.claus->next = $<MC>2.claus;
		    $<MC>1.modu = $<MC>2.modu;
		    }
		 $<MC>$.claus = $<MC>1.claus;
		 $<MC>$.modu = $<MC>1.modu;
		 }
		;

	module :
		BEMOD namen '.' 
		 {
		 modunr		 = ++modcnt;;
		 $<MC>$.claus	 = NULL;
		 $<MC>$.modu	 = init_mod($<N>2);
		 }
		interface strategy body ENMOD '.' 
		 {
		 $<MC>4.modu->interface  = $<I>5;
		 $<MC>4.modu->strategy   = $<C>6;
		 $<MC>4.modu->c_list	 = $<MC>7.claus;
		 $<MC>4.modu->m_list	 = $<MC>7.modu;
		 enter_interface($<MC>4.modu);
		   /* reset modunr and currmodule */
		 modunr		 = $<MC>4.modu->prev->modunr;
		 currmodule	 = $<MC>4.modu->prev;
		 $<MC>$		 = $<MC>4;
		 }
		| clause
		 {
		 $<MC>$.modu	= NULL;
		 $<MC>$.claus 	= $<C>1;
		 }
		;

	namen :
		 { /* empty */
		 $<N>$ = enter_token("default_");
		 }
		| NAME
		 {
		 $<N>$ = $<N>1;
		 }
		;

	interface :
		 { /* empty */
		 $<I>$.export = NULL;
		 $<I>$.import = NULL;
		 $<I>$.global = NULL;
		 }
		| EXPO exports '.' interface
		 {
		 $<I>$.global = $<I>4.global;
		 $<I>$.import = $<I>4.import;
		 for (slp=$<S>2; slp->next; slp=slp->next) ;
		 slp->next = $<I>4.export;
		 $<I>$.export = $<I>2.export;
		 }
		| IMPO NAME CONSTRSEP imports '.' interface
		 {
		 $<I>$.global = $<I>4.global;
		 $<I>$.export = $<I>6.export;
		 $<I>$.import = init_import($<S>4,$<N>2);
		 $<I>$.import->next = $<I>6.import;
		 }
		| GLOB globals '.' interface
		 {
		 $<I>$.export = $<I>4.export;
		 $<I>$.import = $<I>4.import;
		 for (slp=$<S>2; slp->next; slp=slp->next) ;
		 slp->next = $<I>4.global;
		 $<I>$.global = $<S>2;
		 }
		;
	
	strategy :
		stastrat dynstrat 
		 {
		 if (!$<C>1) 
		    $<C>$ = $<C>2;
		 else { /* $<C>$ = $<C>1; */
		    for (currclause=$<C>1; currclause->next; 
			currclause=currclause->next) ;
		    currclause->next = $<C>2;
		    }
		 }
		;

	stastrat :
		 { /* empty */
		 $<C>$ = NULL;
		 }
		| BECOM '.'
		 {
		 aktsign = '-';
		 currclause = newclause();
		 currclause->delpos = COMPILE;
		 currclause->parsed_rule_type = 0;
		 }
		LARR controling ENCOM '.'
		 {
		 aktsign = '+';
		 currclause->p_list = $<P>5;
		 $<C>$ = currclause;
		 }
		;

	dynstrat : 
		 { /* empty */
		 $<C>$ = NULL;
		 }
		| BECON '.' 
		 {
		 aktsign = '-';
		 currclause = newclause();
		 currclause->delpos = CONTROL;
		 currclause->parsed_rule_type = 0;
		 }
		LARR controling ENCON '.'
		 {
		 aktsign = '+';
		 currclause->p_list = $<P>5;
		 $<C>$ = currclause;
		 }
		;

	controling : 
		literal '.'
		| literal '.' controling
		 {
		 $<P>1->next = $<P>3;
		 }
		| literal error controling
		 {
		 yyerrok;
		 $<P>1->next = $<P>3;
		 }
		;

	body :
		modules
		| BEIMP '.' modules ENIMP '.'
		 {
		 $<MC>$ = $<MC>3;
		 }
		;
	
	clause :
		 { /* syntax rule only for semantic actions for/after clause */
		 $<C>$ = currclause = newclause();
		 $<C>$->modul	= currmodule;
		 $<C>$->clnr	= clausenr;
		 $<C>$->lineno	= yylineno;
		 $<C>$->delpos	= PFACT;	/* default */
		 $<C>$->parsed_rule_type = PFACT;
		 $<C>$->poslits  =0;
		 $<C>$->max_termsize  =0;
		 $<C>$->max_termdepth  =0;
		 $<C>$->order 	= clausenr++;	
		 $<C>$->fann	= NULL;
		 $<C>$->next	= NULL;
		 builtnr	= 1;
		 litnr		= 1;
		 poslitnr	= 0;
		 termdepth	=0;
		 termsize	=0;
		 varlist_ptr	= varlist;
		 }
		synclause  
		 {
		 yyerrok;
		 if (litnr > maxlit) maxlit = litnr;
		 if ($<C>1->delpos<0) nofanning = 1;
		 $<C>1->p_list	 = $<P>2;
		 $<P>2->kp	 = $<C>1;
		 $<C>1->nr_built = builtnr-1;
		 $<C>1->nr_lits	 = litnr-1;
		 $<C>1->poslits	 = poslitnr;
		 $<C>1->nrvars	 = varlist_ptr - varlist;

		/* clean up the symbol table */
		while (varlist_ptr > varlist){
			symbtab[*(--varlist_ptr)].index=0;
			}
                 if (litnr <= 2) $<C>1->ddepth = 0;
		 else $<C>1->ddepth = 1;
		 $<C>1->dinf     = litnr-2;

		 /* debugging of termdepth-size */
		 printf("termdepth=%d termsize=%d nrvars=%d\n",
		 	$<C>1->max_termdepth,
		 	$<C>1->max_termsize,
		 	$<C>1->nrvars);
		 }
		opt_constraints  '.'
		 {
		 $<C>1->constr = $<P>4;
		 if (poslitnr > 1) ishorn = 0;
		 }
		 ;

	opt_constraints :
                 { /* empty */
		 $<P>$ = NULL;
                 }
		| CONSTRSEP constraints
		 {
		 $<P>$ = $<P>2;
		 }
		;

	synclause :
                head LARR
                 {
                 aktsign = '-';
                 currclause->delpos = litnr;
		 currclause->parsed_rule_type = 0;
                 }
                tail
                 {
                 aktsign = '+';
                 /* last(head) -> first(tail) */
                 for (pp = $<P>1 ; pp->next ; pp = pp->next) ;
                 pp->next = $<P>4;
                 }
                | head LARR
                 {
                 currclause->delpos = litnr;
		 currclause->parsed_rule_type = 0;
                 }
		| LARR 
		 {
		 currclause->delpos = litnr;
		 currclause->parsed_rule_type = 0;
		 /* default query has index 0 */
		 $<P>$ = init_pred(NULLTERM,0);
		 poslitnr = 1;
		 aktsign = '-';
		 }
		tail 
		 {
		 aktsign = '+';
		 $<P>2->next = $<P>3;
		 $<P>$ = $<P>2;
		 }
		| literal COLMIN 
		 {
		 aktsign = '-';
		 }
		tail 
		 {
		 aktsign = '+';
		 currclause->delpos = PRULE;
		 currclause->parsed_rule_type = PRULE;
		 $<P>1->next = $<P>4;
		 }
		| QUERY
		 {
		 /* default query has index 0 */
		 $<P>$ = init_pred(NULLTERM,0);
		 poslitnr = 1;
		 aktsign = '-';
		 } 
		tail 
		 {
		 aktsign = '+';
		 currclause->delpos = PQUERY;
		 currclause->parsed_rule_type = PQUERY;
		 $<P>2->next = $<P>3;
		 $<P>$ = $<P>2;
		 }
		| literal 
		| error 
		 {
		 yyerrok;
		 $<P>$ = NULL;
		 }
		;

	head :
		literal
		| literal ';' head
		 {
		 $<P>1->next = $<P>3;
		 }
		| literal error head 
		 {
		 yyerrok;
		 $<P>1->next = $<P>3;
		 }
		| literal ';' error 
		;
	
	tail :
		literal
		| literal ',' tail
		 {
		 $<P>1->next = $<P>3;
		 }
		| literal error tail 
		 {
		 yyerrok;
		 $<P>1->next = $<P>3;
		 }
		| literal ',' error
		;

	literal :
		predicate
		 {
		 if ($<P>1->sign == '+') poslitnr += 1;
		 if($<P>1->symb<nrbuilt && $<P>1->sign == '+')
		 	nferror("Illegal usage of a built-in",yylineno);
		 }
		| '~' predicate
		 {
		 $<P>2->sign = (aktsign == '+')? '-' : '+';
		 if($<P>2->symb<nrbuilt && $<P>2->sign == '+')
			nferror("Illegal usage of a built-in",yylineno);
		 if ($<P>2->sign == '+') poslitnr += 1;
		 $<P>$ = $<P>2;
		 }
		| VAR IS numexpr
		 {
		 tp = init_term(NULLTERM,$<N>1,VARIABLE,0,NGROUND);
		 tp->next = $<T>3;
		 $<P>$ = init_pred(tp,IS_SYMB);
		 if (aktsign == '+')
		    nferror("Illegal usage of an assignment",yylineno);

		 VARUSAGE(tp);
		 }
		| glvar ASSIGN term
		 {
		 $<T>1->next = $<T>3;
		 $<P>$ = init_pred($<T>1,ASSI_SYMB);
		 if (aktsign == '+')
		    nferror("Illegal usage of an assignment",yylineno);
		 }
		| glvar EVASSIGN numexpr
		 {
		 $<T>1->next = $<T>3;
		 $<P>$ = init_pred($<T>1,EVAS_SYMB);
		 if (aktsign == '+')
		    nferror("Illegal usage of an assignment",yylineno);
		 }
		| binexpr
		;

	predicate :
		NAME 
		 {
		 $<P>$ = init_pred(NULLTERM,$<N>1);
		 }
		| NAME '(' 
			inctermdepth
			terms ')'
		 {
		 $<P>$ = init_pred($<T>4,$<N>1);
		 /* NOT NECCESSARY
		 if (termdepth > currclause->max_termdepth) 
		 	currclause->termdepth = termdepth;
		*/
		 if (termsize > currclause->max_termsize) 
		 	currclause->max_termsize = termsize;
		 termdepth=0;
		 termsize=0;
		 if ($<N>1 == MKGLOB_SYMB)
		    enter_globarray($<P>$->t_list);
		 }
		;

	exports :
		NAME '/' NUMBER
		 { 
		 enter_name($<N>3,NUMCONST,0,0);
		   /* actual numval is arity in symbtab */
		 $<N>1 = enter_name($<N>1,PREDSYM,numval,modunr);
		 $<S>$= new_symboll($<N>1,NULLLIST);
		 }
		| exports  NAME '/' NUMBER 
		 { /* better left recursion , but reverse order */
		 enter_name($<N>4,NUMCONST,0,0);
		 $<N>2 = enter_name($<N>2,PREDSYM,numval,modunr);
		 $<S>$ = new_symboll($<N>2,$<S>1);
		 }
		| error 
		 {
		 $<S>$ = new_symboll(0,NULLLIST);
		 }
		;

	imports :
		NAME
		 {
		 $<N>1 = enter_name($<N>1,PREDSYM,-1,modunr);
		 $<S>$= new_symboll($<N>1,NULLLIST);
		 }
		| imports NAME 
		 { /* better left recursion , but reverse order */
		 $<N>2 = enter_name($<N>2,PREDSYM,-1,modunr);
		 $<S>$ = new_symboll($<N>2,$<S>1);
		 }
		| error 
		 {
		 $<S>$ = new_symboll(0,NULLLIST);
		 }
		;

	globals :
		GLOBVAR
		 {
		 $<N>1 = enter_globvar($<N>1,currmodule,GLOBAL);
		 $<S>$= new_symboll($<N>1,NULLLIST);
		 }
		| globals GLOBVAR 
		 { 
		 $<N>2 = enter_globvar($<N>2,currmodule,GLOBAL);
		 $<S>$ = new_symboll($<N>2,$<S>1);
		 }
		| error 
		 {
		 $<S>$ = new_symboll(0,NULLLIST);
		 }
		;

	constraints :
		constraint
		| constraint ',' constraints
		 {
		 $<P>1->next = $<P>3;
		 }
		;

	constraint :
		conslist CONSTRAIN conslist
		 {
		 $<T>1->next = $<T>3;
		 $<P>$ = init_pred($<T>1,$<N>2);
		 $<P>$->litnr = ++nr_constraints;
		 $<P>$->claus = currclause;
		 $<P>$->sign = SEM_CONSTR;
		 }
		;

	terms :
		term
		| term ',' terms
		 {
		 $<T>1->next = $<T>3;
		 }
		| term error terms
		 {
		 $<T>1->next = $<T>3;
		 }
		| error
		 {
		 $<T>$ = init_term(NULLTERM,0,CONSTANT,0,GROUND);
		 }
		;

	term :
		VAR
		 { 
		 termsize++;
		 if (currclause->max_termdepth < termdepth)
		 	currclause->max_termdepth = termdepth;
		 $<T>$ = init_term(NULLTERM,$<N>1,VARIABLE,0,NGROUND);
		 VARUSAGE($<T>$);
		 }
		| NUMBER
		 { 
		 termsize++;
		 if (currclause->max_termdepth < termdepth)
		 	currclause->max_termdepth = termdepth;
		 $<T>$ = init_term(NULLTERM,$<N>1,NUMCONST,numval,GROUND);
		 }
		| STRING
		 { 
		 termsize++;
		 if (currclause->max_termdepth < termdepth)
		 	currclause->max_termdepth = termdepth;
		 $<T>$ = init_term(NULLTERM,$<N>1,STRINGCONST,0,GROUND);
		 }
		| glvar
		| STRUCVAR
		 { /* only allowed in constraints => scan.l */
		 $<T>$ = init_term(NULLTERM,$<N>1,SVARIABLE,0,NGROUND);
		 }
		| NAME 
		 { 
		 termsize++;
		 if (currclause->max_termdepth < termdepth)
		 	currclause->max_termdepth = termdepth;
		 $<T>$ = init_term(NULLTERM,$<N>1,CONSTANT,0,GROUND);
		 }
		| NAME '(' inctermdepth terms ')'
		 {
		 termdepth--;
		 termsize++;
		 $<T>$ = init_term($<T>4,$<N>1,CONSTANT,0,GROUND);
		 }
		| '[' inctermdepth list ']'
		 { 
		 termdepth--;
		 termsize++;
		 $<T>$ = init_term($<T>2,HEAD_SYMB,CONSTANT,0,GROUND);
		 }
		| '[' ']'
		 { 
		 termsize++;
		 if (currclause->max_termdepth < termdepth)
		 	currclause->max_termdepth = termdepth;
		 $<T>$ = init_term(NULLTERM,NIL_SYMB,CONSTANT,0,GROUND);
		 }
		;

	list :
		term
		 { /* list ends with nil element */
		 $<T>1->next = init_term(NULLTERM,NIL_SYMB,CONSTANT,0,GROUND);
		 }
		| term ',' list
		 {
		 $<T>1->next = init_term($<T>3,HEAD_SYMB,CONSTANT,0,GROUND);
		 }
		| term '|' term
		 {
		 $<T>1->next = $<T>3;
		 }
		;

	conslist:
		'[' terms ']'
		 {
		 $<T>$ = init_term($<T>2,-1,CONSTSYM,0,GROUND);
		 }
		;

	glvar :
		COPYS
		 {
		 $<T>$ = init_term(NULLTERM,COPI_SYMB,GVARIABLE,0,NGROUND);
		 loc_cpy = 1;
		 }
		| GLOBVAR
		 {
		 termsize++;
		 if (currclause->max_termdepth < termdepth)
		 	currclause->max_termdepth = termdepth;
		 $<T>$ = init_term(NULLTERM,$<N>1,GVARIABLE,0,NGROUND);
		 }
		| GLOBVAR '[' numexpr ']'
		 {
		 termsize++;
		 if (currclause->max_termdepth < termdepth)
		 	currclause->max_termdepth = termdepth;
		 $<T>$ = init_term($<T>3,$<N>1,GVARIABLE,0,NGROUND);
		 }
		;
	
	numexpr :
		NUMBER
		 { 
		 $<T>$ = init_term(NULLTERM,$<N>1,NUMCONST,numval,GROUND);
		 }
		| VAR
		 { 
		 $<T>$ = init_term(NULLTERM,$<N>1,VARIABLE,0,NGROUND);
		 VARUSAGE($<T>$);
		 }
		| glvar
		| '(' numexpr ')'
		 {
		 $<T>$ = $<T>2;
		 }
		| numexpr '+' numexpr
		 {
		 $<T>1->next = $<T>3;
		 $<T>$ = init_term($<T>1,$<N>2,CONSTANT,0,GROUND);
		 }
		| numexpr '-' numexpr
		 {
		 $<T>1->next = $<T>3;
		 $<T>$ = init_term($<T>1,$<N>2,CONSTANT,0,GROUND);
		 }
		| numexpr '*' numexpr
		 {
		 $<T>1->next = $<T>3;
		 $<T>$ = init_term($<T>1,$<N>2,CONSTANT,0,GROUND);
		 }
		| numexpr '/' numexpr
		 {
		 $<T>1->next = $<T>3;
		 $<T>$ = init_term($<T>1,$<N>2,CONSTANT,0,GROUND);
		 }
		| '-' numexpr %prec '*'
		 {  /* dont use $<N>1 == MINUS_SYMB  */
		 $<T>$ = init_term($<T>2,UMI_SYMB,CONSTANT,0,GROUND);
		 }
		| error
		 {
		 $<T>$ = init_term(NULLTERM,0,CONSTANT,0,GROUND);
		 }
		;
	
	binexpr :
		numexpr RELOP numexpr
		 { /*  =<  <  >  >=  */
		 $<T>1->next = $<T>3;
		 $<P>$ = init_pred($<T>1,$<N>2);
		 }
		| term CONSTRAIN term
		 { /* same operator for constraint =/= !! */
		 $<T>1->next = $<T>3;
		 $<P>$ = init_pred($<T>1,NQAL_SYMB);
		 /* NOT NECESSARY
		 if (termdepth > currclause->termdepth) 
		 	currclause->termdepth = termdepth;
		 */
		 if (termsize > currclause->max_termsize) 
		 	currclause->max_termsize = termsize;
		 termdepth=0;
		 termsize=0;
		 }
		| term '=' '=' term
		 {
		 $<T>1->next = $<T>4;
		 $<P>$ = init_pred($<T>1,EQAL_SYMB);
		 /* NOT NECESSARY
		 if (termdepth > currclause->max_termdepth) 
		 	currclause->termdepth = termdepth;
		 */
		 if (termsize > currclause->max_termsize) 
		 	currclause->max_termsize = termsize;
		 termdepth=0;
		 termsize=0;
		 }
		| term '=' term
		 {
		 $<T>1->next = $<T>3;
		 $<P>$ = init_pred($<T>1,UNIF_SYMB);
		 /* NOT NECESSARY
		 if (termdepth > currclause->termdepth) 
		 	currclause->termdepth = termdepth;
		 */
		 if (termsize > currclause->max_termsize) 
		 	currclause->max_termsize = termsize;
		 termdepth=0;
		 termsize=0;
		 }
		;

inctermdepth:
		{ termdepth++;}
%%

#include "lex.yy.c"
#include "init.c"

/***********************************************************************
/* yyerror and additional functions
/* orig. from yaccsed.script
/***********************************************************************/
yyerror(s)  /* Verbesserte Fehlermeldungen */
register char *s; {  /*nach Schreiner*/
extern int yynerrs;
static int list = 0;
if (!list) {
  printf("[error %d] ",yynerrs+1); yywhere();
  if (list=s[strlen(s)-1]==':') fputs(s,stdout);
    else puts(s);
  } else if (*s!='\n') putchar(' '),fputs(s,stdout);
    else putchar('\n'),list=0;
}
extern char yytext[];
extern int yyleng,yylineno;
yywhere() {  /* Fehlerplatzierung durch Bezug auf yytext */
register int colon=0;  /*nach Schreiner*/
if (yylineno>0) {
  if (colon) fputs(", ",stdout);
   printf("line %d",yylineno-((*yytext=='\n')||(!*yytext)));
   colon=1;
  }
switch (*yytext) {
  case '\0':
  case '\n': break;
  default: if (colon) putchar(' ');
    printf("near \"%.*s\"",yyleng>20 ? 20 :
	   yytext[yyleng-1]=='\n' ? yyleng-1 : yyleng,yytext);
    colon=1;
  }
if (colon) fputs(": ",stdout);
}

