/****************************************************
;;; MODULE NAME : defines.h
;;;
;;; PARENT      :
;;;
;;; PROJECT     : MPLOP
;;;
;;; SCCS        : @(#)defines.h	14.1  12 Mar 1996
;;;
;;; AUTHOR      : Janos Breiteneicher
;;;
;;; DATE        : 15.5.91
;;;
;;; DESCRIPTION : header file with all defines
;;;
;;; REMARKS     :
;;;		26.3.93	jsc	increased number of NRLAB
;;;		4.6.93	jsc	new structure for built-ins
;;;		2.7.93	jsc	includes,extern.h
;;;
****************************************************/

#ifndef DEFINES_H
#define DEFINES_H

#include <stdio.h>

char *malloc();
char *strcpy();
char *strncpy();

/*****************************************************/
/* definitions
/*****************************************************/

#define NAMELENGTH 40           /* max. length of a symbol */
#define MAXSYMB    (unsigned )4000         /* length of symbol table */
#define MAXPREDSYMB	10000	/* max. number of predicate symbols */

#define MAXTRAIL   100		/* length of trail for weak unif */
#define NRLAB	   10000	/* max. nr of labels in a term   */

#define MODUSYM 'm'		/* type of symbol is module */
#define PREDSYM 'p'		/* " is a predicate */
#define CONSTANT 'c'		/* " is a constant */
#define NUMCONST 'n'		/* " is a number */
#define STRINGCONST 'z'		/* " is a string */
#define VARIABLE 'v'		/* " is a variable */
#define GVARIABLE 'g'		/* " is a global variable */
#define SVARIABLE 's'		/* " is a structural variable */
#define BUILTIN 'b'		/* " is a built_in   */
#define CONSTSYM 'a'		/* " is a constraint symbol */

#define GROUND	'g'		/* term is ground */
#define NGROUND 'n'		/* comp. term is non-ground */

/*****************************************************/
/* definitions for output (names of predicates & matrix)
/*****************************************************/
#define QUERY_PRTNA 		"query__"
#define LIST_HEAD 		"["
#define ASSIGN_PRTNA		":="
#define EVASSIGN_PRTNA		":is"
#define IS_PRTNA		"is"
#define UNIF_PRTNA		"$unif"
#define EQAL_PRTNA		"$equal"
#define NQAL_PRTNA		"$nequal"
#define GREATER_PRTNA		"$greater"
#define LESS_PRTNA		"$less"
#define GREAEQU_PRTNA		"$geq"
#define LESSEQU_PRTNA		"$leq"

#define NOTEQU_CONSTR		"=/="

#define NIL 			"[]"	/*"nil_"	*/
#define PLUS			"+"
#define MINUS			"-"
#define MULT			"*"
#define DIV			"/"
#define UMINUS			"-_"

/*****************************************************/
/* definitions for fixed inidices in symboltable 
/*****************************************************/
#define HEAD_SYMB		1
#define NIL_SYMB		2
#define PLUS_SYMB		3
#define MINUS_SYMB		4
#define MULT_SYMB		5
#define DIV_SYMB		6
#define UMI_SYMB		7
#define COPI_SYMB		8
#define DEPT_SYMB		9
#define INF_SYMB		10
#define ASSI_SYMB		11
#define EVAS_SYMB		12
#define IS_SYMB			13
#define UNIF_SYMB		14
#define MKGLOB_SYMB		15
#define GREA_SYMB		16
#define LESS_SYMB		17
#define LEQU_SYMB		18
#define GRQU_SYMB		19
#define EQAL_SYMB		20
#define NQAL_SYMB		21

#define VERB(x)		((inwasm_flags.verbose^inwasm_flags.verbose>>x)&1)
#define DEB(x)		((debug^debug>>x)&1)

#define NAM(x)		(symbtab[x].name)

#define INVERSE(x)	(MAXPREDSYMB+x) /* pos symbolindex +MAXPREDSYMB */
#define INV		4	/* log10 MAXPREDSYMB */

/*****************************************************/
/* def's for unify
/*****************************************************/
#define FAIL (-1)
#define SUCCESS 0

/* renaming */
/* PJ
#define NEW_NAME 0
#define OLD_NAME 1
*/
#define ALL 1         /* modes fuer my_rename */
#define ALL_TO_SV 2   /* siehe my_rename(), ~~ NEW_NAME */
#define ONLY_SV 3
#define TO_OLD 4      /*                    ~~ OLD_NAME */
#define SV_TO_V 5
#define SV_TO_XV 6

/* Copy_terms */
#define ONE_TERM 0
#define ALL_TERM 1
#define ALL_ARGS 2

/*****************************************************/
/* definitions for interface
/*****************************************************/
#define GLOBAL 1

/*****************************************************/
/* definitions for fanning clausel in prologstyle
/*****************************************************/
#define PFACT  (-1)
#define PQUERY (-2)
#define PRULE  (-3)
#define AXIOM  (0)
#define REDUCTION  (4)

#define COMPILE (-5)
#define CONTROL (-6)

/*****************************************************/
/* misc def's for code gen
/*****************************************************/
#define IDTAG   0x100000        /* tag for non-num. constants */

#define CONST	0
#define VAR_	1
#define GTERM	2
#define NGTERM	3
#define CRTERM	4
#define EOSTR	5
#define VOID	6
#define CREF	7
#define GLOBO	8
#define NCONST	9
#define SVAR	10

/*****************************************************/
/* constraint types
/*****************************************************/
#define SEM_CONSTR  0
#define TAUT_CONSTR 1
#define SUB_CONSTR  2
#define SYMM_CONSTR 3
#define PUR_CONSTR  4
#define OVER_CONSTR 5

/*****************************************************/
/* definitions (enum type) for built-in types
/*****************************************************/
typedef enum {
	NONE,
	DETERMINISTIC,		/* deterministic built-in	*/
	NONDETERMINISTIC,	/* non- -"-			*/
	INTERNAL,		/* built-in w/internal name	*/
	INFRULE_A,		/* inference rule before or-branch */
	INFRULE_Z,		/* inference rule after or-branch */
	} built_in_type;

#endif
/*######################################################################*/
