
;;;*************************************************************************
;; ---------------------------
;; CONNECTION GRAPH PROCEDURES
;; ---------------------------
(module connection-graph
	(include "setheo.sch")
	(export (make-connection-graph clauses-out clauses-in) (make-eq-links clauses-in))
	(import (globals "globals.scm") (nr-non-horn parser "parser.scm") (unification "unification.scm")))

(define (make-connection-graph clauses-out clauses-in)
;; computes the links for every literal occurrence in the clauses
;; Procedure is destructive in that it modifies the subgoal-links
;; position of each subgoal vector
 (if (pair? clauses-out)
     (begin (make-cg-clause (clause-generic (car clauses-out)) 
                            0 clauses-in)
            (make-connection-graph (cdr clauses-out) clauses-in))))

(define (make-cg-clause generic offset clauses-in)
 (if (vector-ref generic offset)
     (begin 
      (make-cg-subgoal (vector-ref generic offset) clauses-in)
      (set-subgoal-links (vector-ref generic offset)
                   (reverse (subgoal-links (vector-ref generic offset))))
      (if (or (> nr-non-horn 0) foldup)
          (set-subgoal-links (vector-ref generic offset)
               (cons 'reduction (subgoal-links (vector-ref generic offset)))))
      (set-subgoal-links      ; updates links in first clause copy
                   (subgoals-offset-copy generic offset)
                   (subgoal-links (vector-ref generic offset)))
      (make-cg-clause generic (+ 1 offset) clauses-in))))

(define (make-cg-subgoal subgoal clauses-in)
 (if (pair? clauses-in)
     (begin (make-cg-sg-clause subgoal
                  (copy-subgoals (car (clause-copies (car clauses-in))))
		  0 (clause-length (car clauses-in)))
            (make-cg-subgoal subgoal (cdr clauses-in)))))

(define (make-cg-sg-clause subgoal subgoals offset length)
 (if (< offset length)
     (begin (make-cg-sg-sg subgoal (vector-ref subgoals offset) offset)
            (make-cg-sg-clause subgoal subgoals (+ offset 1) length))))

(define (make-cg-sg-sg subgoal other-subgoal offset)
 (if (and (eq? (subgoal-pred subgoal) (subgoal-pred other-subgoal))
          (not (eq? (subgoal-sign subgoal) (subgoal-sign other-subgoal)))
          (or equality
              (unify-lists-stop (subgoal-terms other-subgoal) 
	                        (subgoal-terms subgoal))))
     (begin (or equality (unbind-vars))
            (set-subgoal-links subgoal 
	         (cons (cons (copy-clause (subgoal-copy other-subgoal)) offset)
                       (subgoal-links subgoal))))))

(define (make-eq-links clauses-in)
 (if (pair? clauses-in)
     (begin (make-eq-links-clause 
                  (copy-subgoals (car (clause-copies (car clauses-in))))
		  0 (clause-length (car clauses-in)))
            (make-eq-links (cdr clauses-in)))
     (begin (set! eq-links (reverse eq-links))
            (if (> nr-non-horn 0)
                (set! eq-links (cons 'reduction eq-links))))))

(define (make-eq-links-clause subgoals offset length)
 (if (< offset length)
     (begin (make-eq-links-sg (vector-ref subgoals offset) offset)
            (make-eq-links-clause subgoals (+ offset 1) length))))

(define (make-eq-links-sg subgoal offset)
 (if (and (eq? 'equal (subgoal-pred subgoal)) (eq? '+ (subgoal-sign subgoal)))
     (set! eq-links 
        (cons (cons (copy-clause (subgoal-copy subgoal)) offset) eq-links))))

;(define (insert-reduction clauses)
; (if (pair? clauses)
;     (begin (insert-reduction-literals (clause-generic (car clauses))
;                                       (- (clause-length (car clause)) 1)
;            (insert-reduction (cdr clauses)))))
;
;(define (insert-reduction-literals subgoal-vector index)
; (if 
;  (>= index 0)
;  (begin 
; (if (> (clause-length (caar (subgoal-links (vector-ref subgoal-vector index))
;   (set-subgoal-links (vector-ref subgoal-vector index)
;   (insert-reduction-links (subgoal-links (vector-ref subgoal-vector index))))
;   (insert-reduction-literals subgoal-vector (- index 1)))))
;
;(define (insert-reduction-links links new-links)
; (if (pair? links)
;     (if (= 1 (clause-length (car links))) 
;         (insert-reduction-links (cdr links))
;         (set-cdr! links 
;;;END *********************************************************************
