;;;*************************************************************************
;; ------------------------
;; GENERAL GLOBAL VARIABLES
;; ------------------------
(module globals
	(include "setheo.sch")
	(export trail-root trail v-trail-root v-trail unif-trail-root
		unif-trail c-stack-root c-stack actual-c-stack eq-trail-root
		eq-trail eq-trail-before eq-links difference-counter filename
		formula start-clauses norel global-subgoals actual-copy 
		all-global-subgoals current-subgoal current-complexity
		subgoal-complexity-measure new-complexity current-vars
		new-vars optimized-unification bound-fail actual-depth
		free-sig old-free-sig sig-vector sig-trail sig-trail-position
		old-sig-trail-position proof-output successful-inferences
		successful-matchings successful-reduction-steps inf
		nr-reduction-steps nr-matchings failed-inferences
		constraint-fails total-successful-inferences
		total-successful-matchings total-failed-inferences
		total-constraint-fails index-root index in-index? 
		index-bindings max-variables nth-variable var-assoc-nr?
		constraints regularity regularity-only-for-subgoals sgindex
		nr-subgoals index-size clause-reordering subgoal-alternation
		subgoal-reordering global-reordering clause-type-reliable
		signature-bound equality equal-predecessor-fail depth
		final-depth weighted-depth weighted-depth-entered wd1 wd2 wd3
		matings-pruning test-grounding foldup foldup-root
		strong-regularity index-foldup printindex nr-lemmas lemmatype)
	(export (update-defaults args)))

(define trail-root (list (cons #f #f)))
(define trail trail-root)
(define v-trail-root (list (vector #f #f #f #f)))
(define v-trail v-trail-root)
(define unif-trail-root (list (cons #f #f)))
(define unif-trail unif-trail-root)
(define c-stack-root (list #f))
(define c-stack c-stack-root)
(define actual-c-stack c-stack)
(define eq-trail-root (list #f))
(define eq-trail eq-trail-root)
(define eq-trail-before eq-trail-root)
(define eq-links '())
(define difference-counter 0)
(define filename #f)
(define formula (make-vector 3))
(set-formula-lemmas formula '()) ;; macro
(define start-clauses '())
(define norel #f)
(define global-subgoals #f)
(define actual-copy (make-vector 10))
(define all-global-subgoals #f)
(define current-subgoal #f)
(define current-complexity #f)
(define subgoal-complexity-measure 0)
(define new-complexity #f)
(define current-vars #f)
(define new-vars #f)
(define optimized-unification #f)
(define bound-fail #f)
(define actual-depth 0)
(define free-sig 0)
(define old-free-sig 0)
(define sig-vector #f)
(define sig-trail #f)
(define sig-trail-position 0)
(define old-sig-trail-position 0)
(define proof-output #f)
(define successful-inferences 0)
(define successful-matchings 0)
(define successful-reduction-steps 0)
(define inf 0)
(define nr-reduction-steps 0)
(define nr-matchings 0)
(define failed-inferences 0)
(define constraint-fails 0)
(define total-successful-inferences 0)
(define total-successful-matchings 0)
(define total-failed-inferences 0)
(define total-constraint-fails 0)
(define index-root #f)
(define index #f)
(define in-index? #t)
(define index-bindings (make-vector 100 #f))
(define max-variables 0) 
(define nth-variable 0)
(define var-assoc-nr? #f)
(define constraints #f)
(define regularity #f)
(define regularity-only-for-subgoals #t)
(define sgindex #f)
(define nr-subgoals 0)
(define index-size 0)
(define test-grounding #f)
(define foldup #f)
(define foldup-root (make-vector 10 #f))
(define strong-regularity #f)
(define index-foldup #f)
(define printindex #f)
(define nr-lemmas 0)
(define lemmatype 0)

;;;*************************************************************************
;; -----------------------------------------
;; DEFAULTS SETTINGS FOR THE SEARCH STRATEGY
;; -----------------------------------------

(define clause-reordering #f)
(define subgoal-alternation #f)
(define subgoal-reordering #f)
(define global-reordering #f)
(define clause-type-reliable #t)
(define signature-bound #f)
(define equality #f)
(define equal-predecessor-fail #f)
(define depth 1)
(define final-depth 0)
(define weighted-depth #f)
(define weighted-depth-entered 0)
(define wd1 120)
(define wd2 120)
(define wd3 120)
(define matings-pruning #f)

;;;*************************************************************************
;; ---------------------------
;; UPDATE OF DEFAULTS SETTINGS
;; ---------------------------

(define (update-defaults args)
 (if (pair? (cdr args))
     (cond ((equal? (car args) "-d")
            (set! final-depth (string->integer (cadr args)))
	    (update-defaults (cddr args)))
           ((equal? (car args) "-id")
            (set! depth (string->integer (cadr args)))
	    (update-defaults (cddr args)))
           ((equal? (car args) "-noclreord")
            (set! clause-reordering #f)
	    (update-defaults (cdr args)))
           ((equal? (car args) "-wdr")
            (set! weighted-depth #t)
	    (update-defaults (cdr args)))
           ((equal? (car args) "-wd1")
            (set! weighted-depth #t)
            (set! wd1 (string->integer (cadr args)))
            (update-defaults (cddr args)))
           ((equal? (car args) "-wd2")
            (set! weighted-depth #t)
            (set! wd2 (string->integer (cadr args)))
	    (update-defaults (cddr args)))
           ((equal? (car args) "-wd3")
            (set! weighted-depth #t)
            (set! wd3 (string->integer (cadr args)))
	    (update-defaults (cddr args)))
           ((equal? (car args) "-sig")
            (set! signature-bound #t)
            (set! free-sig (string->integer (cadr args)))
	    (update-defaults (cddr args)))
           ((equal? (car args) "-dynsgreord")
            (set! subgoal-reordering (string->integer (cadr args)))
	    (update-defaults (cddr args)))
           ((equal? (car args) "-complexity")
            (set! subgoal-complexity-measure (string->integer (cadr args)))
	    (update-defaults (cddr args)))
           ((equal? (car args) "-singledelay")
            (set! subgoal-alternation 1)
	    (update-defaults (cdr args)))
           ((equal? (car args) "-reg")
            (set! constraints #t) (set! regularity #t)
	    (update-defaults (cdr args)))
           ((equal? (car args) "-multidelay")
            (set! subgoal-alternation 2)
	    (update-defaults (cdr args)))
           ((equal? (car args) "-globreord")
            (set! global-reordering #t)
	    (update-defaults (cdr args)))
           ((equal? (car args) "-mate")
            (set! matings-pruning #t)
	    (update-defaults (cdr args)))
           ((equal? (car args) "-proof")
            (set! proof-output #t)
	    (update-defaults (cdr args)))
           ((equal? (car args) "-sgindex")
            (set! sgindex #t)
	    (update-defaults (cdr args)))
           ((equal? (car args) "-equality")
            (set! equality #t)
	    (update-defaults (cdr args)))
           ((equal? (car args) "-equality")
            (set! equality #t)
	    (update-defaults (cdr args)))
           ((equal? (car args) "-norel")
            (set! norel #t)
	    (update-defaults (cdr args)))
           ((equal? (car args) "-foldup")
            (set! foldup #t)
	    (update-defaults (cdr args)))
           ((equal? (car args) "-indexfoldup")
            (set! index-foldup #t)
	    (update-defaults (cdr args)))
           ((equal? (car args) "-printindex")
            (set! printindex #t)
	    (update-defaults (cdr args)))
           ((equal? (car args) "-sreg")
            (set! strong-regularity #t) (set! regularity #t)
	    (update-defaults (cdr args)))
           ((equal? (car args) "-fvector")
            (make-vector (string->integer (cadr args)) #f)
	    (update-defaults (cddr args)))
           ((equal? (car args) "-eqpred")
            (let ((option (string->integer (cadr args))))
             (if (= option 0)
                 (begin (set! equal-predecessor-fail 0)
                        (update-defaults (cdr args)))
                 (begin (set! equal-predecessor-fail 1)
                        (update-defaults (cddr args))))))
           ((equal? (car args) "-grounding")
            (set! test-grounding #t)
	    (update-defaults (cdr args)))
	   (else (display "Abort: error in parameter list") 
	         (newline) (exit 1))
)))

;;;*************************************************************************
;; -------------
;; FOR DEBUGGING
;; -------------

(define l1 #f)
(define l2 #f)
(define cl #f)
(define d #f)
(define i 0)
