;;;*************************************************************************
;; --------------------------------------
;; MODEL ELIMINATION PROOF PROCEDURE
;; --------------------------------------
(module solve
	(include "setheo.sch")
	(export ;(solve sg links depth inference nr-sgs)
		(extension sg clause entry depth inference nr-sgs)
		(ref-int int-term) 
		(free-signature-weights sigs) (update-pred-numbers preds))
	(import (globals "globals.scm") (unification "unification.scm")
		(parser "parser.scm") (copy "copy.scm")
		(pruning "pruning.scm") (constraints "constraints.scm")
		(display "display.scm")
		(find-matching-lemma index "index.scm")
		(print-predicate-lemma-indices index "index.scm")
		(comp-in-lemma-index-subgoal index "index.scm")
		(in-lemma-index-subgoal index "index.scm")))

(define (solve sg links depth inference nr-sgs)
 (if 
  (not (null? links))
  (if (and sgindex 
           (begin ; insert unit-lemma if last subgoal
	     (if (= nr-sgs 1)
		 (begin
		   ;(display-subgoal sg) (display " inserted, nr-sgs= ")
		   ;(display nr-sgs) (newline)
		   (set! lemmatype 0)
		   (in-lemma-index-subgoal sg)))
	     #t)
           (find-matching-lemma sg)
	   ;(begin (newline) (display "Subgoal ") (display-subgoal sg)
	;	  (display " solved by index matching") #t)
;	   (set! i (+ 1 i)) #f
	   )
      (let ((old-global-subgoals global-subgoals) (new-inf inference)
            new-sg new-links new-depth)
        (set! successful-inferences (+ successful-inferences 1))
	(set! successful-matchings (+ successful-matchings 1))
	(set! nr-matchings (+ 1 nr-matchings))
;	(next-subgoal-call-fast)   
        (next-subgoal-call new-inf new-sg new-links new-depth 
			 old-global-subgoals (- nr-sgs 1))
	(set! nr-matchings (- nr-matchings 1)))
      (begin 
	(if (eq? (car links) 'reduction)
	    (if foldup
		(reduction-steps sg (subgoal-foldup-predecessor sg)
				 inference nr-sgs)
		(reduction-steps sg (subgoal-predecessor sg)
				 inference nr-sgs))
	    (extension sg (caar links) 
		       (vector-ref 
 			   (copy-subgoals (car (clause-copies (caar links))))
			   (cdar links))
		       depth inference nr-sgs))
	(if subgoal-alternation 
	    (alternate-subgoals sg links depth inference nr-sgs)
	    (solve sg (cdr links) depth inference nr-sgs))))))

(define (reduction-steps sg predsg inference nr-sgs)
 (if 
  predsg
  (begin
   (if 
    (and (not (eq? (subgoal-sign sg) (subgoal-sign predsg)))
         (eq? (subgoal-pred sg) (subgoal-pred predsg))
         (or (not matings-pruning) ; begin of matings pruning
             (> (subgoal-clause-number sg)
                (subgoal-clause-number (subgoal-next predsg)))
             (and (= (subgoal-clause-number sg)
                     (subgoal-clause-number (subgoal-next predsg)))
                  (> (subgoal-number sg)
                     (subgoal-number (subgoal-next predsg))))) ; end mat. pr.
         (or (and (unify-lists-stop (subgoal-terms sg) (subgoal-terms predsg))
	          (or (and (not constraints)
			   (begin
			     (if (or foldup index-foldup)
				 (begin (v-trail-stop)
					(update-time-stamp predsg inference)))
			     #t))
		      (begin (v-trail-stop) ; backtracking marker
			(set! actual-c-stack c-stack)
			(if (propagate-substitution) ; constr. not viol.
			    (begin
			      (if (or foldup index-foldup)
				  (update-time-stamp predsg inference))
			      #t);     below: shallow constraint and unfication
			    (begin (set! c-stack actual-c-stack) ; backtracking
				(undo-v-trail) (unbind-vars) 
				(set! constraint-fails (+ constraint-fails 1))
				#f))))) ; next line ends shallow backtracking
             (begin (set! failed-inferences (+ failed-inferences 1)) #f)))
    (let ((old-global-subgoals global-subgoals) (new-inf inference)
          (old-c-stack actual-c-stack) new-sg new-links new-depth)
     (set! successful-inferences (+ successful-inferences 1))
     (set! successful-reduction-steps (+ successful-reduction-steps 1))
     (set! nr-reduction-steps (+ nr-reduction-steps 1))
;     (newline) (display "Subgoal ") (display-subgoal sg)
;	       (display " solved by reduction")
     (next-subgoal-call new-inf new-sg new-links new-depth
			old-global-subgoals (- nr-sgs 1))
;    (next-subgoal-call-fast)
     ;; deep backtracking:
     (set! nr-reduction-steps (- nr-reduction-steps 1))
     (if (or constraints foldup index-foldup) (undo-v-trail))
     (if constraints (set! c-stack old-c-stack))
     (unbind-vars)))
   (if foldup
       (reduction-steps sg (subgoal-foldup-predecessor predsg)
			inference nr-sgs)
       (reduction-steps sg (subgoal-predecessor predsg)
			inference nr-sgs)))))
     
(define (extension sg clause entry depth inference nr-sgs)
 (if (and (check-bounds clause depth)
          (or (not signature-bound)
              (begin (set! old-free-sig free-sig)
                     (set! old-sig-trail-position sig-trail-position)
                     (if (sig-bound (clause-signature clause))
                         #t
                         (begin (set! free-sig old-free-sig) 
                                (set! bound-fail #t) #f))))
          (or (begin (set! actual-depth depth)
                     (set! actual-copy (car (clause-copies clause)))
		     (not sg)) ; top clause entered: no unification/regularity
              (and (extension-unify-lists-stop (subgoal-terms entry) 
					       (subgoal-terms sg))
		   (or (and (not constraints)
			    (begin (if (or foldup index-foldup)
				       (v-trail-stop)) #t))
		       (begin (v-trail-stop) ; backtracking marker
			(set! actual-c-stack c-stack)
			(if (and (propagate-substitution) ; constr. not viol.
				 (regularity entry (copy-subgoals actual-copy) 
					     sg (clause-length clause)))
			    #t ; below: shallow constraint and unfication
			    (begin (set! c-stack actual-c-stack) ; backtracking
				(undo-v-trail) (unbind-vars) 
				(set! constraint-fails (+ constraint-fails 1))
				#f)))))
	      (begin (if signature-bound ; shallow signature backtracking
			 (begin (untrail-sig old-sig-trail-position)
				(set! free-sig old-free-sig)))
		     #f)))
     (let ((copy actual-copy) (old-global-subgoals global-subgoals)
	   (old-copies (clause-copies clause)) (new-inf inference)
           (old-eq-trail eq-trail-before) 
           new-sg new-links new-depth
           (old-local-free-sig old-free-sig)
           (old-local-sig-trail-position old-sig-trail-position)
           (old-c-stack actual-c-stack))
      (set! successful-inferences (+ successful-inferences 1))
;      (if sg (begin (display "Open depth: ") (display depth)
;                    (display ", inference: ") (display inference)
;		    (display ", clause number: ")
;		    (display (clause-number clause)) (newline)))
      (set-copy-path copy sg)
      (if foldup
	  (if sg
	      (set-subgoal-foldup-predecessors (copy-subgoals copy)
				(- (clause-length (copy-clause copy)) 1) sg) 
	      (set-subgoal-foldup-predecessors (copy-subgoals copy)
        		(- (clause-length (copy-clause copy)) 1) foldup-root)))
;      (display foldup-root) (newline)
      (new-clause-copy clause)
      (set-clause-copies clause (cdr (clause-copies clause)))
;      (set-copy-eq-counter copy difference-counter)
      (set! new-inf (+ difference-counter new-inf (clause-length clause)))
      (if sg (begin (set! new-inf (- new-inf 1))
                    (set-subgoal-next sg entry)))
      (set! inf new-inf)
      (if (or (not sg) (> (clause-length clause) 1) (> difference-counter 0))
          (begin 
            (set-copy-depth copy (- depth 1))
            (set-copy-inference copy new-inf)
            (if weighted-depth
                (set-copy-free-depth100 copy 
                   (- (+ (* 100 (copy-depth copy)) wd1)
                      (* wd2 (+ (clause-length clause)
				difference-counter -1)))))
	    (if (or (not sg) (> (clause-length clause) 1))
	        (begin (vector-set! (copy-subgoals copy) 
			            (clause-length clause) global-subgoals)
                       (set! global-subgoals (copy-subgoals copy))))
            (if (> difference-counter 0)
                (begin (set-subgoal-other-subgoals (car eq-trail) 
                                                   global-subgoals)
                       (set! global-subgoals (copy-eq-trail copy)))
                (set-copy-eq-trail copy #f))
            (if sg 
                (set-subgoal-selected entry #t)
	        (set! all-global-subgoals global-subgoals))))
      (next-subgoal-call new-inf new-sg new-links new-depth
		 old-global-subgoals (+ (- nr-sgs 2) (clause-length clause)))
;      (next-subgoal-call-fast)
      ;; deep backtracking:
      (if sg 
          (begin (set-subgoal-selected entry #f)
                 (if (or constraints foldup index-foldup) (undo-v-trail))
                 (if constraints (set! c-stack old-c-stack))
                 (set! eq-trail old-eq-trail)
                 (unbind-vars)))
      (if signature-bound (begin (untrail-sig old-local-sig-trail-position)
				 (set! free-sig old-local-free-sig)))
      (set-clause-copies clause old-copies))
      ;; end of shallow backtracking:
      (set! failed-inferences (+ failed-inferences 1))))
 
;;;*************************************************************************
;; -----------------
;; SUBGOAL SELECTION
;; -----------------

(define (select-subgoal subgoals index max-index inference)
 (if 
  (= index max-index)
   (if (and subgoal-reordering current-subgoal)
       (if (and global-reordering 
                (vector? (vector-ref subgoals max-index))
                (> (subgoal-depth current-subgoal) 0))
           (select-subgoal (vector-ref subgoals max-index) 0
                (- (vector-length (vector-ref subgoals max-index)) 1)
		inference)
           current-subgoal)
       (begin 
        ; foldup:
        (if (and (or foldup index-foldup)
		 (vector? (vector-ref global-subgoals max-index)))
	    (foldup-subgoal 
	     (subgoal-predecessor (vector-ref global-subgoals 0)) inference))
	(set! global-subgoals (vector-ref subgoals max-index))
	(cond ((vector? global-subgoals)
	       (select-subgoal global-subgoals 0 
                             (- (vector-length global-subgoals) 1)
			     inference))
;              ((pair? global-subgoals) (select-eq-subgoal global-subgoals))
	      (else (display-inferences depth)
	            (newline) (display "     ***** PROOF WITH ") 
		    (display inf) (display " INFERENCES ")
                    (if (> nr-reduction-steps 0) 
                        (begin (display "(") (display nr-reduction-steps) 
                               (display " RED.) ")))
                    (if (> nr-matchings 0) 
                        (begin (display "(") (display nr-matchings) 
                               (display " LEM.) ")))
		    (display "FOUND *****") (newline) (newline)
		    (display "Total successful/failed inferences")
		    (display ": ") 
		    (display total-successful-inferences)
                    (if (> successful-reduction-steps 0)
                        (begin (display " (RED: ") 
		               (display successful-reduction-steps)
                               (display ")")))
                    (if (> total-successful-matchings 0)
                        (begin (display " (LEM: ") 
		               (display total-successful-matchings)
                               (display ")")))
                    (display "/")
		    (display total-failed-inferences) 
;		    (if constraints 
;			(begin (display " (") 
;			       (display total-constraint-fails) (display ")")))
		    (newline)
                    (if proof-output 
                        (begin (newline) (newline) (display "USED CLAUSES")
                           (newline) (display "------------") (newline)
                           (display-proof-clauses (formula-clauses formula))))
                    ;(print-predicate-lemma-indices predicate-list)
		    (newline)
		    (exit 0) #f))))
  (if (subgoal-selected (vector-ref subgoals index))
      (select-subgoal subgoals (+ index 1) max-index inference)
      (if 
        subgoal-reordering
	(begin (compare-subgoals (vector-ref subgoals index))
               (select-subgoal subgoals (+ index 1) max-index inference))
        (vector-ref subgoals index)))))

(define (select-eq-subgoal eq-trail)
 (if (subgoal-selected (car eq-trail))
     (if (subgoal-other-subgoals (car eq-trail)) ;; end of current eq-subgoals
         (if (not current-subgoal) ;; no current eq-subgoal
             (begin 
               (set! global-subgoals (subgoal-other-subgoals (car eq-trail)))
               (if (pair? global-subgoals) 
                   (select-eq-subgoal global-subgoals)
;                  (select-subgoal global-subgoals 0 
;                                   (- (vector-length global-subgoals) 1))
		   ))
             current-subgoal)
         (select-eq-subgoal (cdr eq-trail)))
     (if subgoal-reordering    
         (if current-subgoal 
             (begin (compare-subgoals (car eq-trail))
                    (select-eq-subgoal (cdr eq-trail)))
             (begin (set! current-subgoal (car eq-trail))
                    (select-eq-subgoal (cdr eq-trail))))
         (car eq-trail))))

(define (compare-subgoals subgoal)
 (if current-subgoal
     (if current-complexity 
	 (begin (trail-stop)
;	        (set! new-complexity (termlist-v-complexity 
;		                                 (subgoal-terms subgoal)))
	        (set! new-complexity (subgoal-instantiation subgoal))
		(set! new-vars (unbind-bindings))
		(if (subgoal-conditions subgoal)
		    (begin (set! current-subgoal subgoal)
		           (set! current-vars new-vars)
                           (set! current-complexity new-complexity))))
         (begin (trail-stop)
;		(set! current-complexity (termlist-v-complexity 
;		                           (subgoal-terms current-subgoal)))
	        (set! current-complexity 
                      (subgoal-instantiation current-subgoal))
                (set! current-vars (unbind-bindings))
	        (trail-stop)
;	        (set! new-complexity (termlist-v-complexity 
;		                           (subgoal-terms subgoal)))
	        (set! new-complexity (subgoal-instantiation subgoal))
		(set! new-vars (unbind-bindings))
		(if (subgoal-conditions subgoal)
		    (begin (set! current-subgoal subgoal)
		           (set! current-vars new-vars)
                           (set! current-complexity new-complexity)))))
    (set! current-subgoal subgoal)))

(define (unbind-bindings)
;; makes free all variables back to the previous trailstop, outputs number
;; of modified bindings 
 (if (trail-content)
     (begin (set-car! (caar trail) '()) 
            (trail-previous) (+ 1 (unbind-bindings)))
     (begin (trail-previous) 0)))

(define (ref-int int-term)
;; if int-term is a <bound_variable>, its binding is returned, 
;; otherwise int-term is returned
 (if (or (integer? int-term) (free-variable int-term) 
	 (vector? (car int-term)))
     int-term
     (ref-int (car int-term))))

(define (ref-ext ext-term)
;; if term is a <bound_variable>, its binding is returned, 
;; otherwise term is returned
 (if (or (not ext-term) (non-variable ext-term) (free-variable ext-term))
     ext-term
     (ref-ext (car ext-term))))

;;(define (ref-occ occ-ground-term)
;;;; an occ-ground-term is a bound variable, a term or a nonnegative integer
;;;; if term is a <bound_variable>, its binding is returned, 
;;;; otherwise occ-ground-term is returned
;; (if (or (non-variable occ-ground-term) (free-variable ext-term))
;;     ext-term
;;     (ref-ext1 (car ext-term))))

(define (termlist-var-occ terms)
; counts number of variables in term by setting the variable to
; its current frequency and incrementing this value;
; uses the trail: therefore, before calling main function, say (trail-stop)
 (if (pair? terms)
     (begin (term-var-occ (car terms))
;            (newline) (display "1 :") (set! l1 terms) (repl)
            (termlist-var-occ (cdr terms)))))

(define (term-var-occ term)
; instantiates free variables to 0 and variables with integer bindings n
; to n+1; auxiliary to procedure termlist-var-occ 
 (if (free-variable term) 
     (bind term 0)
     (if (integer? (car term))
         (set-car! term (+ (car term) 1))
         (if (vector? (car term))
             (termlist-var-occ (cdr term))
             (term-var-occ (car term))))))

(define (term-instantiation1 term)
; computes instantiation value of an extended term t as follows:
;           |  ( occ(t) - 1 ) / occ(t)    if t is a variable (occurrence)
; t-i(t) =  {  1                          if t is a constant 
;           |  1/2 + sum(t-i(a_i)/2*n)    if t = f(a_1,...,a_n)
;
 (cond ((integer? (car term))
        (quotient (* 1000 (car term)) (+ (car term) 1)))
       ((constant term) 1000)
       ((vector? (car term))
        (+ 500 (quotient (termlist-instantiation1 (cdr term)) 
                              (* 2 (functor-arity (car term))))))
       (else (term-instantiation1 (car term)))))

(define (termlist-instantiation1 termlist)
 (if (null? (cdr termlist))
     (term-instantiation1 (car termlist))
     (+ (term-instantiation1 (car termlist))
        (termlist-instantiation1 (cdr termlist)))))

(define (termlist-v-complexity terms)
; computes number of symbols in terms - number of distinct variables
; uses the trail: therefore, before calling main function, say (trail-stop)
;                            and afterwards say: (unbind-bindings))
 (if (null? terms)
     0
     (+ (term-v-complexity (ref-ext (car terms))) 
        (termlist-v-complexity (cdr terms)))))

(define (term-v-complexity term)
; computes number of symbols in term - number of distinct variables
; uses the trail: therefore, before calling main function, say (trail-stop)
;                            and afterwards say: (unbind-bindings)
 (cond ((or (not term) (constant term)) 1)
       ((variable term) (bind term #f) 0)
       (else (+ 1 (termlist-v-complexity (cdr term))))))

(define (term-instantiation term)
; computes instantiation value of a term as follows:
;           |  0                          if t is a variable
; t-i(t) =  {  1                          if t is a constant 
;           |  1/2 + sum(t-i(a_i)/2*n)    if t = f(a_1,...,a_n)
;
 (cond ((not term) 0)
       ((variable term) (bind term #f) 0)
       ((constant term) 1000)
       (else (+ 500 (quotient (termlist-instantiation (cdr term)) 
                              (* 2 (functor-arity (car term))))))))

(define (termlist-instantiation termlist)
 (if (null? (cdr termlist))
     (term-instantiation (ref-ext (car termlist)))
     (+ (term-instantiation (ref-ext (car termlist)))
        (termlist-instantiation (cdr termlist)))))

(define (compute-weighted-depth depth free-depth100 entry-inf inf) 
; first line must remain, otherwise segmentation fault 
; (probably because of compiler optimization)
    (quotient (+ 50 free-depth100
                 (quotient (* 100 (- (* depth 100) free-depth100))
                           (+ 100 (* (- inf entry-inf) wd3))))
              100))

;;;*************************************************************************
;; ------------------------
;; COMPUTE AND CHECK BOUNDS
;; ------------------------

(define (free-signature-weights sigs)
 (if (pair? sigs)
     (begin (set-signature-weight (car sigs) 0)
            (free-signature-weights (cdr sigs)))))
 
(define (update-pred-numbers preds)
 (if (pair? preds)
     (begin (set-predsym-number (car preds) 
                         (+ (predsym-number (car preds)) nr-functors))
            (update-pred-numbers (cdr preds)))))
 
(define (sig-bound sigs)
;; the function updates free-sig and the signature vector sig-vector: 
;; free-sig is decremented with the sum of the signature weights of the 
;; new symbols in sigs and the respective positions in the sig-vector are
;; incremented with the signature weights; the new positions in the vector 
;; are remembered in the sig-trail;
;; sig-bound returns #f if the new value of free-sig is < 0
 (if (pair? sigs)
     (if (not (vector-ref sig-vector (signature-number (car sigs))))
         (begin (set! free-sig (- free-sig (signature-weight (car sigs))))
           (if (and (>= free-sig 0) (sig-bound (cdr sigs)))
               (begin (vector-set! sig-vector (signature-number (car sigs))
                                              (signature-weight (car sigs)))
                 (vector-set! sig-trail sig-trail-position 
                                        (signature-number (car sigs)))
                 (set! sig-trail-position (+ sig-trail-position 1)))
               #f))
         (sig-bound (cdr sigs)))
     #t))

(define (untrail-sig old-sig-trail-position)
 (if (> sig-trail-position old-sig-trail-position)
     (begin 
       (set! sig-trail-position (- sig-trail-position 1))
       (vector-set! sig-vector (vector-ref sig-trail sig-trail-position) #f)
       (untrail-sig old-sig-trail-position))))

(define (term-variables termlist)
; extracts set of variables occurring in termlist
; nondestructive, but internally misuses the trail
 (trail-stop)
 (termlist-var-occ termlist)
 (unbind-collect-vars '()))

(define (subgoals-termlists-appended subgoals index)
 (if (< (vector-length subgoals) index)
     (append (subgoal-terms (vector-ref subgoals index))
	     (subgoals-termlists-appended subgoals (+ index 1)))
     '()))

(define (set-subgoal-foldup-predecessors subgoals length predecessor)
 (if (>= length 0)
     (begin (set-subgoal-foldup-predecessor
	         (vector-ref subgoals length) predecessor)
            (set-subgoal-foldup-predecessors
	         subgoals (- length 1) predecessor))))

(define (update-time-stamp sg inference)
 (v-trail-next)
 (vector-set! (car v-trail) 0 sg)
 (vector-set! (car v-trail) 1 9)
 (vector-set! (car v-trail) 2 (subgoal-time-stamp sg))
 (set-subgoal-time-stamp sg inference))

(define (foldup-subgoal subgoal inference) 
 (if foldup
     (let ((c-point (find-c-point subgoal inference)))
	   (v-trail-next) ; trailing:
	   (vector-set! (car v-trail) 0 c-point)
	   (vector-set! (car v-trail) 1 8)
	   (vector-set! (car v-trail) 2
			(subgoal-foldup-predecessor c-point))
	   (v-trail-next)
	   (vector-set! (car v-trail) 0 subgoal)
	   (vector-set! (car v-trail) 1 0)
	   (vector-set! (car v-trail) 2 (subgoal-sign subgoal))
	   (v-trail-next)
	   (vector-set! (car v-trail) 0 subgoal)
	   (vector-set! (car v-trail) 1 8)
	   (vector-set! (car v-trail) 2
			(subgoal-foldup-predecessor subgoal))
	   ; changing sign:
	   (if (eq? (subgoal-sign subgoal) '-)
	       (set-subgoal-sign subgoal '+)
	       (set-subgoal-sign subgoal '-))
	   (set-subgoal-foldup-predecessor subgoal
			       (subgoal-foldup-predecessor c-point))
	   (set-subgoal-foldup-predecessor c-point subgoal)
	   (if (and index-foldup (eq? c-point foldup-root))
	       (begin (set! lemmatype 1) (in-lemma-index-subgoal subgoal))))
     (if (c-point-at-root? subgoal inference)
	 ; since sign of subgoal is not changed
	 (begin (set! lemmatype 1)
		(comp-in-lemma-index-subgoal subgoal)))))

(define (find-c-point subgoal inference)
 (if subgoal
     (if (> (subgoal-time-stamp subgoal) inference) 
         subgoal
	 (find-c-point (subgoal-foldup-predecessor subgoal) inference))
     foldup-root))

(define (c-point-at-root? subgoal inference)
 (if subgoal
     (if (> (subgoal-time-stamp subgoal) inference) 
         #f
	 (c-point-at-root? (subgoal-predecessor subgoal) inference))
     #t))


