;;;*************************************************************************
;;;module name: compiler.sch
;;;author     : Reinhold Letz, TU Munich
;;;date       : 29.10.98
;;;description: compiler for semprop
;;;remarks    : literals are integers /= 0 and may occur multiply in a clause
;;;*************************************************************************

(define (compile) 
 (set! input-vector (make-vector (+ input-size 1) #f))
 (set! cl-length max-cl-length) (set! end-divisor clause-nr)
; (display "end-divisor: ") (display end-divisor) (newline)
 (set! clause-nr 0) (compile-to-semprop)
 (set! unit-vector-length clause-nr)
 (set! unit-vector (make-vector unit-vector-length #f))
 (insert-units-in-vector unit-list)
; (newline) (display-atom-names atom-list)
 (order-atoms)
; (newline) (display-atom-names atom-list)
)

(define (compile-to-semprop) 
 (if (> cl-length 0)
     (begin 
       (cond ((= clause-divisor 0) (set! divisor 1))
             ((= clause-divisor 1) (set! divisor cl-length))
             ((= clause-divisor 2) (set! divisor (* cl-length cl-length)))
             ((= clause-divisor 3) 
              (if (> cl-length 812) ; 812^3 <= greatest-int < 813^3
		  (set! divisor greatest-int)
		  (set! divisor (* cl-length cl-length cl-length))))
             ((= clause-divisor 4)
              (if (> cl-length 152) ; 152^4 <= greatest-int < 153^4  
		  (set! divisor greatest-int)
		  (set! divisor (* cl-length cl-length cl-length cl-length))))
             ((= clause-divisor 5) 
	      (if (> cl-length 55) ; 55^5 <= greatest-int < 56^5
		  (set! divisor greatest-int)
		  (set! divisor	
  		    (* cl-length cl-length cl-length cl-length cl-length))))
             ((= clause-divisor 6) 
	      (if (> cl-length 28) ; 28^6 <= greatest-int < 29^6
		  (set! divisor greatest-int)
		  (set! divisor (* cl-length cl-length cl-length 
				   cl-length cl-length cl-length))))
             (else (newline) (display "Error: wrong clause divisor!") 
                   (exit 1)))
(if (= divisor 0) 
    (set! weight 1)
    (set! weight
	  (max 1 (quotient (quotient greatest-int end-divisor) divisor))))
(vector-set! clauses-weight-vector cl-length weight)
; (newline) (display (vector-ref clauses-weight-vector cl-length))
       (compile-to-semprop-list (vector-ref clauses-vector cl-length))
       (set! cl-length (- cl-length 1))
       (compile-to-semprop))))

(define (compile-to-semprop-list input-clauses) 
 (if (pair? input-clauses)
     (begin (set! clause-nr (+ 1 clause-nr))
            (compile-input-clause (car input-clauses) #f)
            (compile-to-semprop-list (cdr input-clauses)))))

(define (prune-clause)
 (if (element (- (car interm)) (cdr interm))
     (set! tautology #t)
     (if (element (car interm) (cdr interm))
         (begin (set! interm (cdr interm)) (prune-clause))
         (prune-clause1 interm (cdr interm)))))

(define (prune-clause1 precell literals)
 (if (pair? literals)
     (if (element (- (car literals)) (cdr literals))
         (set! tautology #t)
         (if (element (car literals) (cdr literals))
             (begin (set-cdr! precell (cdr literals))
                    (prune-clause1 precell (cdr literals)))
             (begin (set! cl-length (+ cl-length 1))
                    (prune-clause1 literals (cdr literals)))))))

(define (clause-insert)
 (set! tautology #f) (set! cl-length 1) (prune-clause)
   (if (not tautology)
     (begin 
       (if (and split (> cl-length 3)
                (< def-number input-size))
           (split-clause-insert (car interm) (cadr interm) (cddr interm) 
                                cl-length)
           (begin (set! cl-length (min cl-length max-cl-length))
             (vector-set! clauses-vector cl-length
                  (cons interm (vector-ref clauses-vector cl-length)))))))
)

;(define (get-active-clause)
; (if (< literal-index max-cl-length)
;     (get-active-clause-in-list (vector-ref clauses-vector literal-index))
;     #f))
;
;(define (get-active-clause-in-list clauses)
; (if (pair? clauses)
;     (if (c-activating-atom (car clauses))
;         (get-active-clause-in-list (cdr clauses))
;         (begin (set! other-atom (get-literal (c-pos-literals (car clauses))))
;                (if other-atom 
;                    other-atom
;                    (get-literal (c-neg-literals (car clauses))))))
;     (begin (set! literal-index (+ literal-index 1))
;            (get-active-clause))))

(define (split-clause-insert lit1 lit2 lits length)
 (set! def-number (+ def-number 1))
; (newline) (display "def-number: ") (display def-number)
 (vector-set! clauses-vector 3 (cons (list lit1 lit2 def-number)
                                     (vector-ref clauses-vector 3)))
 (vector-set! clauses-vector 2 (cons (list (- lit1) (- def-number))
                                     (cons (list (- lit2) (- def-number))
                                           (vector-ref clauses-vector 2))))
 (set! length (- length 1))
 (if (and (> length 3) (< def-number input-size))
     (split-clause-insert (- def-number) (car lits) (cdr lits) length)
     (vector-set! clauses-vector length (cons (cons (- def-number) lits)
                                      (vector-ref clauses-vector length)))))

;(define (clauses-insert clauses)
; (if (pair? clauses)
;     (begin (set! cl-length (length (car clauses)))
;       (vector-set! clauses-vector cl-length
;             (cons (car clauses) (vector-ref clauses-vector cl-length))))))

(define (element elem list)
 (if (pair? list)
     (if (eq? elem (car list))
         list
         (element elem (cdr list)))
     #f))

(define (compile-input-clause input-clause clause) 
 (set! clause (vector #f cl-length '() '() clause-nr #f))
 (compile-input-literals input-clause clause #f)
 (if (= 1 cl-length) (set! unit-list (cons clause unit-list))))

(define (compile-input-literals literals clause atom)
 (if (pair? literals)
     (begin 
       (if (> (car literals) 0)
           (begin (set! atom (vector-ref input-vector (car literals)))
              (if (not atom) 
                  (begin (set! atom (make-atom (car literals)))
                         (if (> (car literals) hash-size)
                             (set! def-list (cons atom def-list))
                             (set! atom-list (cons atom atom-list)))
                         (set! nr-atoms (+ 1 nr-atoms))
                         (vector-set! input-vector (car literals) atom))) 
              (set-a-pos-clauses atom (cons clause (a-pos-clauses atom)))
              (set-a-pos-frequency atom (+ weight (a-pos-frequency atom)))
              (set-c-pos-literals clause (cons atom (c-pos-literals clause))))
           (begin (set! atom (vector-ref input-vector (- (car literals))))
              (if (not atom) 
                  (begin (set! atom (make-atom (- (car literals))))
                         (if (> (- (car literals)) hash-size)
                             (set! def-list (cons atom def-list))
                             (set! atom-list (cons atom atom-list)))
                         (set! nr-atoms (+ 1 nr-atoms))
                         (vector-set! input-vector (- (car literals)) atom)))
              (set-a-neg-clauses atom (cons clause (a-neg-clauses atom)))
              (set-a-neg-frequency atom (+ weight (a-neg-frequency atom)))
              (set-c-neg-literals clause (cons atom (c-neg-literals clause)))))
       (compile-input-literals (cdr literals) clause #f))))

(define (insert-units-in-vector unit-list)
 (if (pair? unit-list)
     (begin (set! unit-index (+ unit-index 1))
            (vector-set! unit-vector unit-index (car unit-list))
            (insert-units-in-vector (cdr unit-list)))))

(define (order-atoms)
 (end-order-atoms atom-list)
 (if split 
     (begin (end-order-atoms def-list)
            (set-cdr! (last def-list) atom-list)
            (set! atom-list def-list)))
 (if display-atomlist (begin (display-atom-names atom-list) (newline)))
 (set! atom-vector (list->vector atom-list))
; (begin (set! pre-atom-vector (make-vector (+ frequency-size 1) '()))
;        (pre-order-atoms atom-list)
;        (set! atom-vector (make-vector nr-atoms))
;        (fill-atom-vector (vector-ref pre-atom-vector 1) 1 (- nr-atoms 1)))
)

;(define (pre-order-atoms atoms)
; (if (pair? atoms)
;     (begin (if max-frequency
;                (set-a-frequency (car atoms) 
;                          (quotient (max (a-pos-frequency (car atoms)) 
;                                         (a-neg-frequency (car atoms)))
;                                    end-divisor))
;                (set-a-frequency (car atoms) 
;                          (quotient (+ (a-pos-frequency (car atoms)) 
;                                       (a-neg-frequency (car atoms)))
;                                    end-divisor)))
;      (set-a-frequency (car atoms)
;                       (min frequency-size (a-frequency (car atoms))))
;      (vector-set! pre-atom-vector (a-frequency (car atoms))
;         (cons (car atoms) 
;               (vector-ref pre-atom-vector (a-frequency (car atoms)))))
;      (pre-order-atoms (cdr atoms)))))
;
;(define (fill-atom-vector atoms pre-index atom-index)
; (cond ((pair? atoms)
;        (vector-set! atom-vector atom-index (car atoms))
;        (fill-atom-vector (cdr atoms) pre-index (- atom-index 1)))
;       ((>= atom-index 0) (set! pre-index (+ pre-index 1))
;        (fill-atom-vector (vector-ref pre-atom-vector pre-index) 
;                          pre-index atom-index))))

(define (end-order-atoms atoms)
 (let (max-atom max-atom-list)
  (cond ((pair? atoms) (set! max-atom-list (max-atom atoms))
               (set! max-atom (car max-atom-list))
         (cond ((eq? max-atom (car atoms)) '())
               (else (set-car! max-atom-list (car atoms))
                  (set-car! atoms max-atom)))
         (end-order-atoms (cdr atoms))))))

(define (max-atom-atoms atoms atom-pred max min) 
 (let (pos neg new-atom)
  (if 
   (pair? atoms)
    (begin (set! new-atom (car atoms))
      (set! pos (a-pos-frequency new-atom))
      (set! neg (a-neg-frequency new-atom))
      (if max-frequency (set! pos (+ pos min)))
    (cond 
     ((>= pos neg)
      (cond 
       ((> max pos) (max-atom-atoms (cdr atoms) atom-pred max min))
       ((= max pos)
        (cond ((>= min neg)
               (max-atom-atoms (cdr atoms) atom-pred max min))
              (else (max-atom-atoms (cdr atoms) atoms pos neg))))
       (else (max-atom-atoms (cdr atoms) atoms pos neg))))
     ;; in the following: neg > pos:
     ((> max neg) (max-atom-atoms (cdr atoms) atom-pred max min))
     ((= max neg)
      (cond ((>= min pos)
             (max-atom-atoms (cdr atoms) atom-pred max min))
            (else (max-atom-atoms (cdr atoms) atoms neg pos))))
     (else (max-atom-atoms (cdr atoms) atoms neg pos))))
   atom-pred)))

(define (hash-key value) (obj->0..2^x-1 value 16))

(define (hash-reset) (vector-fill! hash-table #f))

(define (hash-insert-key! index obj)
  (if (or (eq? index 0) (eq? index hash-size))
      (set! index 1))
  (if (and (vector-ref hash-table index) 
	   (not (equal?  (vector-ref hash-table index) obj)))
      (hash-insert-key! (+ hash-step index) obj)
      (begin (vector-set! hash-table index obj)
	     index)))

(define (hash-member-key? index obj)
  (if (or (eq? index 0) (eq? index hash-size))
      (set! index 1))
  (if (not (vector-ref hash-table index))
      #f
      (if (equal? (vector-ref hash-table index) obj)
	  index
	  (hash-member-key? (hash-step  index) obj))))

(define (hashify-poslits pos-lits neg-lits)
  (if (null? pos-lits)
      (hashify-neglits neg-lits)   
      (cons (hash-insert! (car pos-lits)) 
            (hashify-poslits (cdr pos-lits) neg-lits))))

(define (hashify-neglits neg-lits)
  (if (null? neg-lits)
      '()
      (cons (- (hash-insert! (car neg-lits))) 
            (hashify-neglits (cdr neg-lits)))))
