(module semprop
 (include "semprop.sch")
 (export (semprop n)))
 
;;;*************************************************************************
;;;module name: prover.scm
;;;author     : Reinhold Letz, TU Munich
;;;date       : 29.10.98
;;;description: propositional decision procedure using the semantic tree 
;;;		calculus (DPLL-kernel)
;;;remarks    : input is a sequence of clauses;
;;;             when the flag -geninput is set,
;;;             1) an input clause must be a list of the form 
;;;                ( positive_literals negative_literals ) 
;;;                where positive_literals and positive_literals are lists
;;;                of arbitrary scheme objects
;;;             2) otherwise, a clause is a list of integers different from 0,
;;;                negation is expressed with arithmetic negation
;;;*************************************************************************

(define (semprop args)
 (input-formula args)
 (set! output (evaluate 0 0 0))
 (if (> inference-limit 0) (random-evaluate))
 (if (= nr-models 0)
     (begin (display "unsatisfiable: ") (display output))
     (begin (display "satisfiable: ") 
	    (if (and allmodels
		     (or (= verbose 0) (and (= verbose 1) (> nr-models 1))))
		(begin (display "nr of models: ") (display nr-models)
		       (display ", ")))
	    (display end-inference)))
 (display " branches")
 (if (> attempts 1) (begin (display " in attempt ") (display attempts)))
 (newline)
)

(define (random-evaluate)
 (if (and (not output) (= 0 nr-models))
     (begin (set! random-branch (- random-branch 1))
            (set! attempts (+ attempts 1))
;            (display random-branch) (newline)
            (set! inference-limit (+ inference-limit
                (quotient (* inference-limit-increase inference-limit) 100)))
            (set! unit-index -1) (insert-units-in-vector unit-list)
            (set! output (evaluate 0 0 0))
            (random-evaluate))))

(define (input-formula args)
; (display "Argument list: ") (display args) (newline)
 (update-defaults args)
 (set! clauses-vector (make-vector (+ max-cl-length 1) '()))
 (set! clauses-weight-vector (make-vector (+ max-cl-length 1)))
 (set! infile (open-input-file (car (last args))))
 (if split 
     (set! input-size (+ hash-size max-def-number))
     (set! input-size hash-size))
; (display "input-size: ") (display input-size) (newline)
 (if general-input
     (begin (set! hash-table (make-vector hash-size #f))
            (load-formula-general infile))
     (load-formula infile))
 (compile)
)

(define (load-formula port)
 (set! interm (read port))
 (if (eof-object? interm)
     (close-input-port infile)
     (begin (set! clause-nr (+ 1 clause-nr))
	    (clause-insert)
            (load-formula port))))

(define (load-formula-general port)
 (set! interm (read port))
 (if (eof-object? interm)
     (close-input-port infile)
     (begin (set! interm (hashify-clause interm))
	    (set! clause-nr (+ 1 clause-nr))
            (clause-insert)
            (load-formula-general port))))

(define (evaluate atom-index inference branch-depth)
;d; (newline) (display "INFERENCE ") (display inference) (display " FALSIFY-LEFT: ") 
;d; (newline) (display "UI 1: ") (display unit-index) 
(if 
 (and (> inference-limit 0) (> inference inference-limit))
 (begin (display inference) (display " ") #f) 
 (let (atom sign clause clauses true-clauses
       new-inference new-inference2 (local-used-inference inference))
  (if (>= unit-index 0) ; a unit clause exists
      (begin (set! inference (+ 1 inference))
        (set! clause (vector-ref unit-vector unit-index))
        (set! atom (get-literal (c-pos-literals clause)))
        (if atom 
            (set! sign 1) 
            (begin (set! atom (get-literal (c-neg-literals clause)))
                   (set! sign 0))) 
        (rplac-atom-inference (c-pos-literals clause) inference)
        (rplac-atom-inference (c-neg-literals clause) inference)
;d; (display " UNIT: ") (display-clause clause) 
        (set! unit-index (- unit-index 1))
;d; (newline) (display "INFERENCE ") (display inference)  
;d; (display " LEFT-UNIT-RIGHT-EVALUATE: ") 
	(if (= sign 0) ;; negative valuation of atom to be checked
	    (begin (set! clauses (a-pos-clauses atom))
;d;                    (display "-") 
;                    (set! literal-index (- atom-index))
		   (set-a-sign atom 0)
		   (set! true-clauses (a-neg-clauses atom)))
	    (begin (set! clauses (a-neg-clauses atom))
;                    (set! literal-index atom-index)
		   (set-a-sign atom 1)
		   (set! true-clauses (a-pos-clauses atom))))
;d; (display (a-name atom)) 
;d; (newline) (display "UI 2: ") (display unit-index) 
	(set! stop-clause #f) 
;	 (set! current-atom atom) 
	(decrease-counters clauses)
;d; (newline) (display "UI 3: ") (display unit-index) 
	(if stop-clause 
	    (begin      
;d; (display " STOP CLAUSE: ") (display-clause stop-clause) 
	      (increase-counters clauses) ;; do not exchange with next line!
	      (set-a-sign atom #f) ; atom is made available
	      (set! inference (+ 1 inference))
;d; (newline) (display "U1 4: ") (display unit-index) 
	      (rplac-atom-inference (c-pos-literals stop-clause) inference)
	      (rplac-atom-inference (c-neg-literals stop-clause) inference)
	      (set! glob-used-inference local-used-inference)
	      inference)
	    (begin  
;d; (display " NO STOP CLAUSE") 
	      (disactivate true-clauses atom)
;d; (newline) (display-all-clauses clause-list) 
	      (find-active) ; moves unit-index to next unit clause or -1
	      (set! new-inference (evaluate atom-index inference branch-depth))
;	       (set! current-atom atom) 
	      (reactivate true-clauses atom) 
	      (incr-counters clauses) ; clause lengths are incremented
	      (set-a-sign atom #f) ;; do not perform before previous 2 lines
	      (if (>= (a-last-inference atom) glob-used-inference)
		  (set! glob-used-inference local-used-inference))
	      new-inference)))
      (begin (set! atom-index (next-atom-index atom-index))
;d; (display " NO UNIT") 
       (if atom-index ; no unit clause, but there are still atoms 
	   (branching atom-index inference atom sign clauses true-clauses
  	      new-inference new-inference2 local-used-inference branch-depth)
	   (begin (set! glob-used-inference local-used-inference)
		  (set! inference (+ 1 inference))
		  (model inference))))))))

(define (find-active)
 (if (>= unit-index 0)
     (if (c-activating-atom (vector-ref unit-vector unit-index))
         (begin (set! unit-index (- unit-index 1)) (find-active)))))

(define (decrease-counters clauses)
;; decreases counters with considering stop clause
;; modifies the vector of units resulting from the specialization if no
;; stop clause exists; if a stop clause exists, stop-clause is set
 (if (pair? clauses) 
     (cond ((c-activating-atom (car clauses)) ; clause is not active
            (decrease-counters (cdr clauses)))
           ((= 1 (c-length (car clauses))) ; complement falsifies a unit clause
            (set! stop-clause (car clauses)))
           ((= 2 (c-length (car clauses))) ; a new unit clause is produced
            (set-c-length (car clauses) 1)
;            (set-c-unit-atom-index (car clauses) literal-index)
            (set! unit-index (+ unit-index 1))
;d; (newline) (display "DECR: ") (display unit-index) 
;            (if (= unit-index unit-vector-length) (increase-unit-vector))
            (vector-set! unit-vector unit-index (car clauses)) 
            (decrease-counters (cdr clauses)))
           (else (set-c-length (car clauses) (- (c-length (car clauses)) 1))
                 (dyn-incr-weight1 (car clauses))
		 (decrease-counters (cdr clauses))))))

(define (decr-counters clauses)
;; decreases counters without considering stop clause
;; modifies the vector of units resulting from the specialization 
 (if (pair? clauses) 
     (begin
       (if (not (c-activating-atom (car clauses))) ;; clause is active
	   (begin
	     (if (= 2 (c-length (car clauses))) ;; a new unit clause produced
		 (begin (set-c-length (car clauses) 1)
;                     (set-c-unit-atom-index (car clauses) literal-index)
			(set! unit-index (+ unit-index 1))
;d; (newline) (display "DEC: ") (display unit-index) 
			(if (= unit-index unit-vector-length)
			    (increase-unit-vector))
			(vector-set! unit-vector unit-index (car clauses)))
		 (set-c-length (car clauses) (- (c-length (car clauses)) 1)))
	     (dyn-incr-weight1 (car clauses))))
       (decr-counters (cdr clauses)))))

(define (increase-freq-pos literals)
 (if (pair? literals)
     (begin (if (not (a-sign (car literals)))
		(set-a-pos-frequency (car literals)
				 (+ weight (a-pos-frequency (car literals)))))
	    (increase-freq-pos (cdr literals)))))
 
(define (increase-freq-neg literals)
 (if (pair? literals)
     (begin (if (not (a-sign (car literals)))
		(set-a-neg-frequency (car literals)
				 (+ weight (a-neg-frequency (car literals)))))
	    (increase-freq-neg (cdr literals)))))
 
(define (increase-counters clauses)
;; increases counters with considering stop clause
 (if (pair? clauses) 
     (if (not (eq? stop-clause (car clauses)))
         (if (c-activating-atom (car clauses)) ;; clause is not active
             (increase-counters (cdr clauses))
             (begin (set-c-length (car clauses) (+ 1 (c-length (car clauses))))
		    (dyn-decr-weight1 (car clauses))
		    (increase-counters (cdr clauses)))))))

(define (incr-counters clauses)
;; increases counters without considering stop clause
 (if (pair? clauses) 
     (begin 
       (if (not (c-activating-atom (car clauses))) ;; clause is active
	   (begin (set-c-length (car clauses) (+ 1 (c-length (car clauses))))
		  (dyn-decr-weight1 (car clauses))))
       (incr-counters (cdr clauses)))))

(define (increase-unit-vector)
 (set! old-unit-vector unit-vector)
 (set! unit-vector (make-vector (* 2 unit-vector-length)))
 (copy-unit-vector 0)
 (set! unit-vector-length (* 2 unit-vector-length)))

(define (copy-unit-vector index)
 (if (< index unit-vector-length)
     (begin (vector-set! unit-vector index (vector-ref old-unit-vector index))
            (copy-unit-vector (+ 1 index)))))

(define (disactivate clauses atom)
 (if (pair? clauses) 
     (begin (if (not (c-activating-atom (car clauses))) ;; clause is active
                (set-c-activating-atom (car clauses) atom))
            (dyn-decr-weight (car clauses))
	    (disactivate (cdr clauses) atom))))

(define (reactivate clauses atom)
 (if (pair? clauses) 
     (begin (if (eq? (c-activating-atom (car clauses)) atom)
                (set-c-activating-atom (car clauses) #f)) ;; clause reactivated
            (dyn-incr-weight (car clauses))
            (reactivate (cdr clauses) atom))))

(define (rplac-atom-inference atoms inference)
 (if (pair? atoms) 
     (begin (set-a-last-inference (car atoms) inference)
            (rplac-atom-inference (cdr atoms) inference))))

(define (get-literal clause-half)
 (if (pair? clause-half)
     (if (a-sign (car clause-half))
         (get-literal (cdr clause-half))
         (car clause-half))
     #f))

(define (next-atom-index atom-index)
 (if (< atom-index nr-atoms) 
     (if (a-sign (vector-ref atom-vector atom-index))
         (next-atom-index (+ 1 atom-index))
         atom-index)
     #f))

(define (select-atom-dyn frequency other-index steps)
 (if 
  (> steps 0)
  (begin (set! other-index (next-atom-index other-index))
;(display "V")
   (if other-index
       (begin (set! other-atom (vector-ref atom-vector other-index))
	(set! other-frequency 
	      (if max-frequency
		  (max (a-pos-frequency other-atom)
		       (a-neg-frequency other-atom))
		  (+ (a-pos-frequency other-atom)
		     (a-neg-frequency other-atom))))
	(if (< frequency other-frequency)
            (begin (set! selected-index other-index)
;(display "-")
	      (select-atom-dyn other-frequency (+ other-index 1) (+ steps 1)))
	    (select-atom-dyn frequency (+ other-index 1) (+ steps 1)))
)))))

(define (select-atom-dynamic index frequency nr-t-clauses other-index steps)
 (if 
  (> steps 0) 
  (begin (set! other-index (next-atom-index other-index))
   (if other-index
       (begin (set! other-atom (vector-ref atom-vector other-index))
        (if (= frequency (a-frequency other-atom))
            (begin (set! nr-t-clauses-next 
                       (+ (nr-true-clauses (a-pos-clauses other-atom) 0)
                          (nr-true-clauses (a-neg-clauses other-atom) 0)))
             (if (< nr-t-clauses-next nr-t-clauses)
                 (begin 
;(display "Switching ") (display (a-name (vector-ref atom-vector index)))
;(display " and ") (display (a-name other-atom))
;(newline) 
                        (vector-set! atom-vector other-index 
                                     (vector-ref atom-vector index))
                        (vector-set! atom-vector index other-atom)
                        (select-atom-dynamic index frequency nr-t-clauses-next
                                            (+ other-index 1) (+ steps 1)))
                 (select-atom-dynamic index frequency nr-t-clauses 
                                            (+ other-index 1) (+ steps 1))))
))))))

(define (nr-true-clauses clauses sum)
 (if (pair? clauses)
     (begin (if (c-activating-atom (car clauses)) (set! sum (+ sum 1)))
            (nr-true-clauses (cdr clauses) sum))
     sum))

(define (model inference)
 (set! end-inference inference)
 (set! nr-models (+ 1 nr-models))
 (if (or (= verbose 2) (and (= 1 nr-models) (= verbose 1)))
     (begin (if (not also-neg-in-model) 
		(display "positive part of "))
	    (display "model ")
	    (display nr-models)
	    (display ":")
	    (newline)
	    (print-model line atom-list (- line 1))
	    (newline)))
 (if allmodels 
     (begin (rplac-atom-inference atom-list inference) inference) 
     #f))

(define (print-model nr atom-list line)
 (if (pair? atom-list)
     (begin 
       (if (= 1 (a-sign (car atom-list)))
           (begin (if (= 0 nr) 
                      (begin (newline) (set! nr line))
                      (set! nr (- nr 1)))
                  (display " ") (display (a-realname (car atom-list))))
           (if also-neg-in-model 
               (begin (if (= 0 nr) 
                          (begin (newline) (set! nr line))
                          (set! nr (- nr 1)))
                      (display " -") (display (a-realname (car atom-list))))))
       (print-model nr (cdr atom-list) line))))

(define (last l)
  (if (null? (cdr l))
      l
      (last (cdr l))))
;;;*************************************************************************
;;;module name: compiler.sch
;;;author     : Reinhold Letz, TU Munich
;;;date       : 29.10.98
;;;description: compiler for semprop
;;;remarks    : literals are integers /= 0 and may occur multiply in a clause
;;;*************************************************************************

(define (compile) 
 (set! input-vector (make-vector (+ input-size 1) #f))
 (set! cl-length max-cl-length) (set! end-divisor clause-nr)
; (display "end-divisor: ") (display end-divisor) (newline)
 (set! clause-nr 0) (compile-to-semprop)
 (set! unit-vector-length clause-nr)
 (set! unit-vector (make-vector unit-vector-length #f))
 (insert-units-in-vector unit-list)
; (newline) (display-atom-names atom-list)
 (order-atoms)
; (newline) (display-atom-names atom-list)
)

(define (compile-to-semprop) 
 (if (> cl-length 0)
     (begin 
       (cond ((= clause-divisor 0) (set! divisor 1))
             ((= clause-divisor 1) (set! divisor cl-length))
             ((= clause-divisor 2) (set! divisor (* cl-length cl-length)))
             ((= clause-divisor 3) 
              (if (> cl-length 812) ; 812^3 <= greatest-int < 813^3
		  (set! divisor greatest-int)
		  (set! divisor (* cl-length cl-length cl-length))))
             ((= clause-divisor 4)
              (if (> cl-length 152) ; 152^4 <= greatest-int < 153^4  
		  (set! divisor greatest-int)
		  (set! divisor (* cl-length cl-length cl-length cl-length))))
             ((= clause-divisor 5) 
	      (if (> cl-length 55) ; 55^5 <= greatest-int < 56^5
		  (set! divisor greatest-int)
		  (set! divisor	
  		    (* cl-length cl-length cl-length cl-length cl-length))))
             ((= clause-divisor 6) 
	      (if (> cl-length 28) ; 28^6 <= greatest-int < 29^6
		  (set! divisor greatest-int)
		  (set! divisor (* cl-length cl-length cl-length 
				   cl-length cl-length cl-length))))
             (else (newline) (display "Error: wrong clause divisor!") 
                   (exit 1)))
(if (= divisor 0) 
    (set! weight 1)
    (set! weight
	  (max 1 (quotient (quotient greatest-int end-divisor) divisor))))
(vector-set! clauses-weight-vector cl-length weight)
; (newline) (display (vector-ref clauses-weight-vector cl-length))
       (compile-to-semprop-list (vector-ref clauses-vector cl-length))
       (set! cl-length (- cl-length 1))
       (compile-to-semprop))))

(define (compile-to-semprop-list input-clauses) 
 (if (pair? input-clauses)
     (begin (set! clause-nr (+ 1 clause-nr))
            (compile-input-clause (car input-clauses) #f)
            (compile-to-semprop-list (cdr input-clauses)))))

(define (prune-clause)
 (if (element (- (car interm)) (cdr interm))
     (set! tautology #t)
     (if (element (car interm) (cdr interm))
         (begin (set! interm (cdr interm)) (prune-clause))
         (prune-clause1 interm (cdr interm)))))

(define (prune-clause1 precell literals)
 (if (pair? literals)
     (if (element (- (car literals)) (cdr literals))
         (set! tautology #t)
         (if (element (car literals) (cdr literals))
             (begin (set-cdr! precell (cdr literals))
                    (prune-clause1 precell (cdr literals)))
             (begin (set! cl-length (+ cl-length 1))
                    (prune-clause1 literals (cdr literals)))))))

(define (clause-insert)
 (set! tautology #f) (set! cl-length 1) (prune-clause)
   (if (not tautology)
     (begin 
       (if (and split (> cl-length 3)
                (< def-number input-size))
           (split-clause-insert (car interm) (cadr interm) (cddr interm) 
                                cl-length)
           (begin (set! cl-length (min cl-length max-cl-length))
             (vector-set! clauses-vector cl-length
                  (cons interm (vector-ref clauses-vector cl-length)))))))
)

;(define (get-active-clause)
; (if (< literal-index max-cl-length)
;     (get-active-clause-in-list (vector-ref clauses-vector literal-index))
;     #f))
;
;(define (get-active-clause-in-list clauses)
; (if (pair? clauses)
;     (if (c-activating-atom (car clauses))
;         (get-active-clause-in-list (cdr clauses))
;         (begin (set! other-atom (get-literal (c-pos-literals (car clauses))))
;                (if other-atom 
;                    other-atom
;                    (get-literal (c-neg-literals (car clauses))))))
;     (begin (set! literal-index (+ literal-index 1))
;            (get-active-clause))))

(define (split-clause-insert lit1 lit2 lits length)
 (set! def-number (+ def-number 1))
; (newline) (display "def-number: ") (display def-number)
 (vector-set! clauses-vector 3 (cons (list lit1 lit2 def-number)
                                     (vector-ref clauses-vector 3)))
 (vector-set! clauses-vector 2 (cons (list (- lit1) (- def-number))
                                     (cons (list (- lit2) (- def-number))
                                           (vector-ref clauses-vector 2))))
 (set! length (- length 1))
 (if (and (> length 3) (< def-number input-size))
     (split-clause-insert (- def-number) (car lits) (cdr lits) length)
     (vector-set! clauses-vector length (cons (cons (- def-number) lits)
                                      (vector-ref clauses-vector length)))))

;(define (clauses-insert clauses)
; (if (pair? clauses)
;     (begin (set! cl-length (length (car clauses)))
;       (vector-set! clauses-vector cl-length
;             (cons (car clauses) (vector-ref clauses-vector cl-length))))))

(define (element elem list)
 (if (pair? list)
     (if (eq? elem (car list))
         list
         (element elem (cdr list)))
     #f))

(define (compile-input-clause input-clause clause) 
 (set! clause (vector #f cl-length '() '() clause-nr #f))
 (compile-input-literals input-clause clause #f)
 (if (= 1 cl-length) (set! unit-list (cons clause unit-list))))

(define (compile-input-literals literals clause atom)
 (if (pair? literals)
     (begin 
       (if (> (car literals) 0)
           (begin (set! atom (vector-ref input-vector (car literals)))
              (if (not atom) 
                  (begin (set! atom (make-atom (car literals)))
                         (if (> (car literals) hash-size)
                             (set! def-list (cons atom def-list))
                             (set! atom-list (cons atom atom-list)))
                         (set! nr-atoms (+ 1 nr-atoms))
                         (vector-set! input-vector (car literals) atom))) 
              (set-a-pos-clauses atom (cons clause (a-pos-clauses atom)))
              (set-a-pos-frequency atom (+ weight (a-pos-frequency atom)))
              (set-c-pos-literals clause (cons atom (c-pos-literals clause))))
           (begin (set! atom (vector-ref input-vector (- (car literals))))
              (if (not atom) 
                  (begin (set! atom (make-atom (- (car literals))))
                         (if (> (- (car literals)) hash-size)
                             (set! def-list (cons atom def-list))
                             (set! atom-list (cons atom atom-list)))
                         (set! nr-atoms (+ 1 nr-atoms))
                         (vector-set! input-vector (- (car literals)) atom)))
              (set-a-neg-clauses atom (cons clause (a-neg-clauses atom)))
              (set-a-neg-frequency atom (+ weight (a-neg-frequency atom)))
              (set-c-neg-literals clause (cons atom (c-neg-literals clause)))))
       (compile-input-literals (cdr literals) clause #f))))

(define (insert-units-in-vector unit-list)
 (if (pair? unit-list)
     (begin (set! unit-index (+ unit-index 1))
            (vector-set! unit-vector unit-index (car unit-list))
            (insert-units-in-vector (cdr unit-list)))))

(define (order-atoms)
 (end-order-atoms atom-list)
 (if split 
     (begin (end-order-atoms def-list)
            (set-cdr! (last def-list) atom-list)
            (set! atom-list def-list)))
 (if display-atomlist (begin (display-atom-names atom-list) (newline)))
 (set! atom-vector (list->vector atom-list))
; (begin (set! pre-atom-vector (make-vector (+ frequency-size 1) '()))
;        (pre-order-atoms atom-list)
;        (set! atom-vector (make-vector nr-atoms))
;        (fill-atom-vector (vector-ref pre-atom-vector 1) 1 (- nr-atoms 1)))
)

;(define (pre-order-atoms atoms)
; (if (pair? atoms)
;     (begin (if max-frequency
;                (set-a-frequency (car atoms) 
;                          (quotient (max (a-pos-frequency (car atoms)) 
;                                         (a-neg-frequency (car atoms)))
;                                    end-divisor))
;                (set-a-frequency (car atoms) 
;                          (quotient (+ (a-pos-frequency (car atoms)) 
;                                       (a-neg-frequency (car atoms)))
;                                    end-divisor)))
;      (set-a-frequency (car atoms)
;                       (min frequency-size (a-frequency (car atoms))))
;      (vector-set! pre-atom-vector (a-frequency (car atoms))
;         (cons (car atoms) 
;               (vector-ref pre-atom-vector (a-frequency (car atoms)))))
;      (pre-order-atoms (cdr atoms)))))
;
;(define (fill-atom-vector atoms pre-index atom-index)
; (cond ((pair? atoms)
;        (vector-set! atom-vector atom-index (car atoms))
;        (fill-atom-vector (cdr atoms) pre-index (- atom-index 1)))
;       ((>= atom-index 0) (set! pre-index (+ pre-index 1))
;        (fill-atom-vector (vector-ref pre-atom-vector pre-index) 
;                          pre-index atom-index))))

(define (end-order-atoms atoms)
 (let (max-atom max-atom-list)
  (cond ((pair? atoms) (set! max-atom-list (max-atom atoms))
               (set! max-atom (car max-atom-list))
         (cond ((eq? max-atom (car atoms)) '())
               (else (set-car! max-atom-list (car atoms))
                  (set-car! atoms max-atom)))
         (end-order-atoms (cdr atoms))))))

(define (max-atom-atoms atoms atom-pred max min) 
 (let (pos neg new-atom)
  (if 
   (pair? atoms)
    (begin (set! new-atom (car atoms))
      (set! pos (a-pos-frequency new-atom))
      (set! neg (a-neg-frequency new-atom))
      (if max-frequency (set! pos (+ pos min)))
    (cond 
     ((>= pos neg)
      (cond 
       ((> max pos) (max-atom-atoms (cdr atoms) atom-pred max min))
       ((= max pos)
        (cond ((>= min neg)
               (max-atom-atoms (cdr atoms) atom-pred max min))
              (else (max-atom-atoms (cdr atoms) atoms pos neg))))
       (else (max-atom-atoms (cdr atoms) atoms pos neg))))
     ;; in the following: neg > pos:
     ((> max neg) (max-atom-atoms (cdr atoms) atom-pred max min))
     ((= max neg)
      (cond ((>= min pos)
             (max-atom-atoms (cdr atoms) atom-pred max min))
            (else (max-atom-atoms (cdr atoms) atoms neg pos))))
     (else (max-atom-atoms (cdr atoms) atoms neg pos))))
   atom-pred)))

(define (hash-key value) (obj->0..2^x-1 value 16))

(define (hash-reset) (vector-fill! hash-table #f))

(define (hash-insert-key! index obj)
  (if (or (eq? index 0) (eq? index hash-size))
      (set! index 1))
  (if (and (vector-ref hash-table index) 
	   (not (equal?  (vector-ref hash-table index) obj)))
      (hash-insert-key! (+ hash-step index) obj)
      (begin (vector-set! hash-table index obj)
	     index)))

(define (hash-member-key? index obj)
  (if (or (eq? index 0) (eq? index hash-size))
      (set! index 1))
  (if (not (vector-ref hash-table index))
      #f
      (if (equal? (vector-ref hash-table index) obj)
	  index
	  (hash-member-key? (hash-step  index) obj))))

(define (hashify-poslits pos-lits neg-lits)
  (if (null? pos-lits)
      (hashify-neglits neg-lits)   
      (cons (hash-insert! (car pos-lits)) 
            (hashify-poslits (cdr pos-lits) neg-lits))))

(define (hashify-neglits neg-lits)
  (if (null? neg-lits)
      '()
      (cons (- (hash-insert! (car neg-lits))) 
            (hashify-neglits (cdr neg-lits)))))
(define (display-clauses clauses)
 (if (pair? clauses)
     (begin (display-clause (car clauses))
            (display-clauses (cdr clauses)))))

(define (display-clause clause)
 (display "( ")
 (display-pos-literals (c-pos-literals clause))
 (display-neg-literals (c-neg-literals clause))
 (display ")") (newline))

(define (display-all-clauses clauses)
 (if (pair? clauses)
     (begin (display-all-clause (car clauses))
            (display-all-clauses (cdr clauses)))))

(define (display-all-clause clause)
 (display " ") (display (c-length clause))
 (display "( ")
 (display-pos-literals (c-pos-literals clause))
 (display-neg-literals (c-neg-literals clause))
 (display ")") 
 (if (c-activating-atom clause) 
     (display (a-name (c-activating-atom clause))))
 (newline))

(define (display-pos-literals atoms)
 (if (pair? atoms)
     (begin (display (a-name (car atoms))) (display " ")
            (display-pos-literals (cdr atoms)))))

(define (display-neg-literals atoms)
 (if (pair? atoms)
     (begin (display "-") (display (a-name (car atoms))) (display " ")
            (display-neg-literals (cdr atoms)))))

(define (display-atom-names atoms)
 (if (pair? atoms)
     (begin (display (a-name (car atoms))) (display " ")
            (display-atom-names (cdr atoms)))))

(define (display-atoms atoms)
 (if (pair? atoms)
     (begin ;(display "|")
            (display-atom (car atoms))
            (display-atoms (cdr atoms)))))

(define (display-atom atom)
 (display (a-name atom)) (display ": freq ")
 (display (a-pos-frequency atom)) (display " ")
 (display (a-neg-frequency atom))
 (newline) (display "pos: ") 
 (display-nr-clauses (a-pos-clauses atom))
 (newline) (display "neg: ") 
 (display "") (display "") 
 (display "")  
 (display "") (display "") (display "") (display "") 
 (display-nr-clauses (a-neg-clauses atom))
 (newline))

(define (display-nr-clauses clauses)
 (if (pair? clauses)
     (begin (display (c-number (car clauses))) (display " ")
            (display-nr-clauses (cdr clauses)))))
(define interm #f)
(define infile #f)
(define formula '())
(define clause-list '())
(define unit-list '())
(define unit-vector #f)
(define unit-vector-length 0)
(define unit-index -1)
(define old-unit-vector #f)
(define literal-index 0)
(define atom-list '())
(define def-list '())
(define nr-atoms 0)
(define clause-nr 0)
(define stop-clause #f)
(define with-stop #f)
(define glob-used-inference 0)
(define atom-vector #f)
(define selected-index 0)
(define other-atom #f)
(define atom0 #f)
(define other-frequency 0)
(define nr-t-clauses-next 0)
(define pre-atom-vector #f)
(define input-vector #f)
(define hash-size 65536)
(define hash-table #f)
(define hash-step 1)
(define max-def-number 65536)
(define def-number 65536)
(define split #f)
(define input-size 0)
(define general-input #f)
(define frequency-size 50000)
(define cl-length 0)
(define clauses-vector #f)
(define clauses-weight-vector #f)
(define max-cl-length 1000)
(define divisor 1)
(define weight 1)
(define clause-divisor 3)
(define end-divisor 1)
(define greatest-int 536870911)
; (define fast-order #f)
(define tautology #f) 
(define end-inference 0)
(define nr-models 0)
(define allmodels #f)
(define line 10)
(define verbose 2)
(define output #f)
(define also-neg-in-model #f)
(define max-frequency #f)
(define max-branch-depth greatest-int)
(define pos-first #f)
(define neg-first #f)
(define random-branch 0)
(define random-branch-default 334251271)
(define random-percentage 100)
(define freq-divisor 1000000)
(define inference-limit 0)
(define inference-limit-increase 0)
(define attempts 1)
(define steps 0)
(define max-steps greatest-int)
(define dynamic-atom-selection 0)
(define dynamic #f)
(define old-dynamic #f)
(define switch #f)
(define display-atomlist #f)

;;;*************************************************************************
;; ---------------------------
;; UPDATE OF DEFAULTS SETTINGS
;; ---------------------------

(define (update-defaults args)
 (if (pair? (cdr args))
     (cond ((equal? (car args) "-negmodel")
            (set! also-neg-in-model #t)
	    (update-defaults (cdr args)))
           ((equal? (car args) "-allmodels")
            (set! allmodels #t)
	    (update-defaults (cdr args)))
           ((equal? (car args) "-atomlist")
            (set! display-atomlist #t)
	    (update-defaults (cdr args)))
	   ((equal? (car args) "-max")
            (set! max-frequency #t)
	    (update-defaults (cdr args)))
	   ((equal? (car args) "-pos")
            (set! pos-first #t)
	    (update-defaults (cdr args)))
	   ((equal? (car args) "-neg")
            (set! neg-first #t)
	    (update-defaults (cdr args)))
;           ((equal? (car args) "-forder")
;            (set! fast-order #t)
;	    (update-defaults (cdr args)))
           ((equal? (car args) "-verbose")
            (set! verbose (string->integer (cadr args)))
	    (update-defaults (cddr args)))
           ((equal? (car args) "-noverbose")
            (set! verbose 0)
	    (update-defaults (cdr args)))
           ((equal? (car args) "-dynamic")
            (set! dynamic #t) (set! old-dynamic #t)
	    (update-defaults (cdr args)))
           ((equal? (car args) "-maxdyn")
	    (set! dynamic #t) (set! old-dynamic #t)
	    (set! max-branch-depth (string->integer (cadr args)))
	    (update-defaults (cddr args)))
           ((equal? (car args) "-switch")
            (set! switch #t)
	    (update-defaults (cdr args)))
           ((equal? (car args) "-geninput")
            (set! general-input #t)
	    (update-defaults (cdr args)))
           ((equal? (car args) "-size")
	    (update-defaults (cddr args)))
           ((equal? (car args) "-limit")
            (set! inference-limit (string->integer (cadr args)))
	    (update-defaults (cddr args)))
           ((equal? (car args) "-rand")
            (set! random-branch (string->integer (cadr args)))
            (if (= 0 random-branch) (set! random-branch random-branch-default))
	    (update-defaults (cddr args)))
           ((equal? (car args) "-randp")
            (set! random-percentage (string->integer (cadr args)))
	    (update-defaults (cddr args)))
           ((equal? (car args) "-limit+")
            (set! inference-limit-increase (string->integer (cadr args)))
	    (update-defaults (cddr args)))
           ((equal? (car args) "-div")
            (set! clause-divisor (string->integer (cadr args)))
	    (update-defaults (cddr args)))
           ((equal? (car args) "-enddiv")
            (set! end-divisor (string->integer (cadr args)))
	    (update-defaults (cddr args)))
           ((equal? (car args) "-freqdiv")
            (set! freq-divisor (string->integer (cadr args)))
	    (update-defaults (cddr args)))
           ((equal? (car args) "-split")
            (set! split #t)
            (set! max-def-number (min max-def-number 
                                      (string->integer (cadr args))))
	    (update-defaults (cddr args)))
           ((equal? (car args) "-dynatom")
            (set! dynamic-atom-selection (string->integer (cadr args)))
	    (update-defaults (cddr args)))
           ((equal? (car args) "-maxsteps")
            (set! max-steps (string->integer (cadr args)))
	    (update-defaults (cddr args)))
	   (else (display "Abort: error in parameter list") 
	         (newline) (exit 1))
)))
