%% This file is in -*- prolog -*- mode
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%                                                                           %%
%%      Version:  1.00   Date: 23/06/98   File: mod_isol.pl                  %%
%% Last Version:                          File:                              %%
%% Changes:                                                                  %%
%% 23/06/98 Created                                                          %%
%%                                                                           %%
%% Purpose: This filter performs some basic simplifying for a certain class  %%
%%          of problems. If a conjecture c contains a literal p that has only%%
%%          one connection  --p to the rest of the formula (p and --p can be %%
%%          unified), then p is replaced by the rest of the clause containing%%
%%          --p and the unifier of p and --p is applied to the newly formed  %%
%%          c'.                                                              %%
%%          This filter is activated by the switch                           %%
%%              -isol                                                        %%
%%          REMARK: This is the worst stuff I had to program in years, this  %%
%%                  may show in the code.                                    %%
%%                                                                           %%
%% Author:  Gernot Stenz                                                     %%
%%                                                                           %%
%% Usage:   prolog mod_isol.pl                                               %%
%%                                                                           %%
%%                                                                           %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%                                                                           %%
%% Dynamic predicates                                                        %%
%%                                                                           %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%                                                                           %%
%% Imports and Exports                                                       %%
%%                                                                           %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- set_flag( occur_check, on ).

%*****************************************************************************
%* process_switch_ground( +ArrangedSwitch )
%*****************************************************************************
%* - what is to be done if -ground is found
%*****************************************************************************

process_switch_isol( ["-isol"] ) :-
	cmd_arg(tptpout,_),
	assert( cmd_arg(isol,[]) ),
	writeln( '% *** option: isol' ),
        !.

process_switch_isol( ["-isol"] ) :-
	assert( cmd_arg(isol,[]) ),
	writeln( '# *** option: isol' ).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%                                                                           %%
%% Local predicates                                                          %%
%%                                                                           %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%
%% isol( problemwise, +ClauseList, +SwitchList, -NewClauseList )
%%           
%% PURPOSE: Perform simplifying for isolated literals.
%%          
%%          
%%          
%% EXAMPLES: 
%%           
%% REMARKS: 
%%          
%% Author: Gernot Stenz	Created 23/06/98
%% Last Change: 
%%              
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

isol( problemwise, ClauseList, _, NewClauseList ) :-
	cmd_arg( isol, _ ),
	%% length( ClauseList, L ),
	%% write( "Simplifying list of "), write(L), writeln(" clauses..." ),
	isol1( ClauseList, 1, NewClauseList1 ),
	%% writeln( "Removing pure clauses..." ),
	remove_pure_clauses( NewClauseList1, NewClauseList1, NewClauseList ),
	!.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%
%% isol1()
%%           
%% PURPOSE: 
%%          
%%          
%%          
%% EXAMPLES: 
%%           
%% REMARKS: 
%%          
%% Author: Gernot Stenz	Created 23/06/98
%% Last Change: 
%%              
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% terminate the recursion
isol1( ClauseList, Pos, ClauseList ) :- 
	length( ClauseList, L ),
	Pos > L,
	!.

%% the conjecture case
isol1(  OldClauseList, Pos, NewClauseList ) :-
	%% write("Checking clause "), write( Pos ), nl,
	my_nth1( Pos, OldClauseList, Clause ),
	Clause = clause( _Name, conjecture, Literals, _Constraints, _Dict), 
	%% do for all literals in Clause...
	check_literals_for_connections( Literals, 1, Clause, Pos, OldClauseList, NewClauseList1 ),
	Pos1 is Pos+1,
	isol1(  NewClauseList1, Pos1, NewClauseList ),
	!.

%% all other cases
isol1(  OldClauseList, Pos, NewClauseList ) :-
	Pos1 is Pos+1,
	isol1(  OldClauseList, Pos1, NewClauseList ),
	!.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%
%% check_literals_for_connections()
%%           
%% PURPOSE: 
%%          
%%          
%%          
%% EXAMPLES: 
%%           
%% REMARKS: 
%%          
%% Author: Gernot Stenz	Created 23/06/98
%% Last Change: 
%%              
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

check_literals_for_connections( Literals, LPos, _, _, ClauseList, ClauseList ) :- 
	length( Literals, L ),
	LPos > L,
	%% writeln("Done checking Litlist..."),
	!.

check_literals_for_connections( Literals, LitPos, Clause, ClausePos, OldClauseList, NewClauseList ) :-
	my_nth1( LitPos, Literals, Lit ),	
	%% write("Checking Literal "), write(Lit), write(" at position "), 
	%% write( LitPos ), nl,
	connections_with_clauses( Lit, Clause, OldClauseList, ConnectedClauses ),
	length( ConnectedClauses, Connections ),
	( Connections == 0 ->
	    %% no connections at all, this clause is pure. we can stop here,
	    %% the clause will be deleted later on.
	    NewClauseList = OldClauseList
	;
	( Connections == 1 ->
	    %% Just one connection... we can simplify here, provided the connection is
	    %% to a different clause
	    %% extract the clause
	    ConnectedClauses = [ [CClause, CPos] ],
	    ( Clause == CClause ->
		%% Uh oh, connection to the same clause, so clause is pure
		%% see above
		NewClauseList = OldClauseList
	    ;
	        %% nl, writeln( "Applying Simplification!!!"), nl,
	        %% split up the connected clause
		CClause =.. [ _, _, _, CLits, CConsts, CDict ],
		%% split up the conjecture clause
		Clause =.. [ clause, Name, conjecture, _, Consts, Dict ],
		%% prepare the actual simplifying
		extract_from_list( CPos, CLits, CLit, CLow, CHigh ),
		append( CLow, CHigh, ReducedCLits ),
		extract_from_list( LitPos, Literals, Lit, Low, High ),
		%% make a new copy to keep the connected clause untouched
		copy_term( [ReducedCLits,CLit], [ReducedCLits1,CLit1] ),
		%% write( "Connection: "), write( Lit ), 
		%% write( " and " ), writeln( CLit ),
		%% apply the unifier
		Lit =.. [_, Atom],
		CLit1 =.. [_, CAtom ],
		%% Occur check not necessary after copy_term...
		Atom = CAtom,               
		%% build the new literal list
		append( Low, ReducedCLits1, Intermediate ),
		append( Intermediate, High, NewLits ),
		%% build the new constraints and dict
		append( Consts, CConsts, NewConsts ),
		append( Dict, CDict, NewDict ),
		%% build the new clause
		NewClause =.. [ clause, Name, conjecture, NewLits, NewConsts, NewDict ],
		%% replace Clause with NewClause in ClauseList
		extract_from_list( ClausePos, OldClauseList, _, LowClauses, HighClauses ),
		append( LowClauses, [ NewClause | HighClauses ], NewClauseList1 ),
		%% recursion proceeds from the same literal position
		check_literals_for_connections( NewLits, LitPos, NewClause, ClausePos, NewClauseList1, NewClauseList )
	    )
	;
	    %% else there is naught to be done, just go on with the next literal
	    LitPos1 is LitPos+1,
	    check_literals_for_connections( Literals, LitPos1, Clause, ClausePos, OldClauseList, NewClauseList )
	)),
	!.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%
%% connections_with_clauses( +Lit, +Clause, +ClauseList, -ConnectedClauses )
%%           
%% PURPOSE: 
%%          
%%          
%%          
%% EXAMPLES: 
%%           
%% REMARKS: 
%%          
%% Author: Gernot Stenz	Created 23/06/98
%% Last Change: 
%%              
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

connections_with_clauses( _, _, [], [] ) :- !.

connections_with_clauses( Lit, Clause, [FirstClause|RestClauses], ConnectedClauses ) :-
	( Clause == FirstClause ->
	    %% we are comparing with literals from the same clause,
	    %% so we have to make a copy of the literal to enable
	    %% finding a connection
	    copy_term( Lit, Lit1 )
	;
	    Lit1 = Lit
	),
	FirstClause =.. [ clause, _, _, Lits, _, _ ],
	connections_with_lit( Lit1, Lits, FirstClause, ConnectedClauses1 ),
	connections_with_clauses( Lit, Clause, RestClauses, ConnectedClauses2 ),
	append( ConnectedClauses1, ConnectedClauses2, ConnectedClauses ),
	!.

%% deal with comments et al.
connections_with_clauses( Lit, Clause, [_|RestClauses], ConnectedClauses ) :-
	connections_with_clauses( Lit, Clause, RestClauses, ConnectedClauses ),
	!.

connections_with_lit( _, [], _, [] ) :- !.

connections_with_lit( Lit, [FirstLit|RestLits], FirstClause, ConnectedClauses ) :-
	%% write( "Literal: " ), writeln( Lit ),
	( var( Lit ) ->
	    %% writeln( "Literal is variable!!!" ),
	    true
	;
	    true
	),
	Lit =.. [ Sign, Atom ],
	FirstLit =.. [ FirstSign, FirstAtom ],
	comp_sign( Sign, FirstSign ),
	%% the following contortions are inevitable as eclipse does not feature
	%% unify/2.
	set_flag( occur_check, on ),
	\+ \+ Atom = FirstAtom,
	set_flag( occur_check, off ),
	connections_with_lit( Lit, RestLits, FirstClause, ConnectedClauses1 ),
	FirstClause =.. [ _, _, _, FirstLits, _, _ ],
	position( FirstPos, FirstLits, FirstLit ),
	ConnectedClauses = [ [FirstClause,FirstPos]| ConnectedClauses1 ],
	!.

%% once we reach this clause it is clear there is no connection with FirstLit
connections_with_lit( Lit, [_|RestLits], FirstClause, ConnectedClauses ) :-
	connections_with_lit( Lit, RestLits, FirstClause, ConnectedClauses ),
	!.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%
%% remove_pure_clauses()
%%           
%% PURPOSE: 
%%          
%%          
%%          
%% EXAMPLES: 
%%           
%% REMARKS: 
%%          
%% Author: Gernot Stenz	Created 23/06/98
%% Last Change: 
%%              
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

remove_pure_clauses( [], _, [] ) :- !.

remove_pure_clauses( [FirstClause|RestClauses], ClauseList, NewClauseList ) :-
	%% position( P, ClauseList, FirstClause ),
	%% write( "Checking clause "), write(P), writeln( " for purity..."),
	FirstClause =.. [ clause, _, _, Literals, _, _ ],
	( connections_of_literals(Literals, FirstClause, ClauseList) ->
	    remove_pure_clauses( RestClauses, ClauseList, NewClauseList1 ),
	    NewClauseList = [ FirstClause | NewClauseList1 ]
	;
	    %% writeln("Removing pure clause."),
	    remove_pure_clauses( RestClauses, ClauseList, NewClauseList )
	),
	!.
	
remove_pure_clauses( [FirstClause|RestClauses], ClauseList, [FirstClause|NewClauseList] ) :-
	remove_pure_clauses( RestClauses, ClauseList, NewClauseList ),
	!.


connections_of_literals([], _, _) :-
	%% writeln("All literals are externally connected!"),
	!.

connections_of_literals( [Lit|RestLits], Clause, ClauseList) :-
	connections_with_clauses( Lit, Clause, ClauseList, ConnectedClauses ),
	( external_connections( Clause, ConnectedClauses ) ->
	    connections_of_literals( RestLits, Clause, ClauseList)
	;
	    fail
	),
	!.

external_connections( _, [] ) :- !, 
	%% writeln("No external connection!"),
	fail.

external_connections( Clause, [ [CClause,_] | RestConnections ] ) :-
	( Clause == CClause ->
	    external_connections( Clause, RestConnections )
	;
	    %% writeln("External connection found!"),
	    %% writeln( "Clauses:"), writeln( Clause ), writeln( CClause ), nl,
	    true
	),
	!.


comp_sign( ++, -- ).

comp_sign( --, ++ ).


my_nth1( _, [], _ ) :- !.

my_nth1( 1, [ Element | _ ], Element ) :- !.

my_nth1( Pos, [ _ | Rest ], Element ) :-
	Pos1 is Pos-1,
	my_nth1( Pos1, Rest, Element ),
	!.

extract_from_list( _, [], _, [], [] ) :- !.

extract_from_list( 1, [Fma|High], Fma, [], High ) :- !.

extract_from_list( Index, [First|Rest], Fma, [First|Low], High ) :-
	Index1 is Index-1,
	extract_from_list( Index1, Rest, Fma, Low, High ),
	!.

position( 0, [], _ ) :- !.

position( 1, [First|_], Elem ) :- 
	First == Elem,
	!.

position( Pos, [_|Rest], Elem ) :- 
	position( Pos1, Rest, Elem ),
	( Pos1 > 0 ->
	    Pos is Pos1+1
	;
	    Pos is 0
	),
	!.

:- set_flag( occur_check, off ).